"""Nodes used in defining an SQL query plan."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Generic, Mapping, Optional, Sequence, Tuple

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId, DagNode, DisplayedProperty, MetricFlowDag
from metricflow_semantics.sql.sql_exprs import SqlExpressionNode
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.sql.sql_table import SqlTable
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import Self, override

logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=False)
class SqlPlanNode(DagNode["SqlPlanNode"], ABC):
    """A node in the SQL plan model.

    In that model:
    * A source node that takes data into the graph e.g. SQL tables or SQL queries in a literal string format.
    * SELECT queries can be modeled as nodes that transform the data.
    * Statements like ALTER TABLE don't fit well, but they could be modeled as just a single sink node.
    * SQL queries in where conditions could be modeled as another SqlQueryPlan.
    * SqlRenderableNode() indicates nodes where plan generation can begin. Generally, this will be all nodes except
      the SqlTableNode() since my_table.my_column wouldn't be a valid SQL query.

    Is there an existing library that can do this?
    """

    @abstractmethod
    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        raise NotImplementedError

    @property
    @abstractmethod
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        """If possible, return this as a select statement node."""
        raise NotImplementedError

    @property
    @abstractmethod
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        """If possible, return this as SQL table node."""
        raise NotImplementedError

    @abstractmethod
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        """Return the SELECT columns that are in this node or the closest `SqlSelectStatementNode` of its ancestors.

        * For a SELECT statement node, this is just the columns in the node.
        * For a node that has a SELECT statement node as its only parent (e.g. CREATE TABLE ... AS SELECT ...), this
          would be the SELECT columns in the parent.
        * If not known (e.g. an arbitrary SQL statement as a string), return None.
        * This is used to figure out which columns are needed at a leaf node of the DAG for column pruning.
        * A SQL table could refer to a CTE, so a mapping from the name of the CTE to the CTE node should be provided to
          get the associated SELECT columns.
        """
        raise NotImplementedError

    @abstractmethod
    def copy(self) -> Self:
        """Create a shallow copy of this node."""
        raise NotImplementedError


class SqlPlanNodeVisitor(Generic[VisitorOutputT], ABC):
    """An object that can be used to visit the nodes of an SQL plan.

    See similar visitor DataflowPlanVisitor.
    """

    @abstractmethod
    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_table_node(self, node: SqlTableNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_cte_node(self, node: SqlCteNode) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError


@dataclass(frozen=True)
class SqlSelectColumn:
    """Represents a column in the select clause of an SQL query."""

    expr: SqlExpressionNode
    # Always require a column alias for simplicity.
    column_alias: str


@dataclass(frozen=True)
class SqlJoinDescription:
    """Describes how sources should be joined together."""

    # The source that goes on the right side of the JOIN keyword.
    right_source: SqlPlanNode
    right_source_alias: str
    join_type: SqlJoinType
    on_condition: Optional[SqlExpressionNode] = None

    def with_right_source(self, new_right_source: SqlPlanNode) -> SqlJoinDescription:
        """Return a copy of this but with the right source replaced."""
        return SqlJoinDescription(
            right_source=new_right_source,
            right_source_alias=self.right_source_alias,
            join_type=self.join_type,
            on_condition=self.on_condition,
        )


@dataclass(frozen=True)
class SqlOrderByDescription:  # noqa: D101
    expr: SqlExpressionNode
    desc: bool


@dataclass(frozen=True, eq=False)
class SqlSelectStatementNode(SqlPlanNode):
    """Represents an SQL Select statement.

    Attributes:
        select_columns: The columns to select.
        from_source: The source of the data for the select statement.
        from_source_alias: Alias for the from source.
        join_descs: Descriptions of the joins to perform.
        group_bys: The columns to group by.
        order_bys: The columns to order by.
        where: The where clause expression.
        limit: The limit of the number of rows to return.
        distinct: Whether the select statement should return distinct rows.
    """

    _description: str
    select_columns: Tuple[SqlSelectColumn, ...]
    from_source: SqlPlanNode
    from_source_alias: str
    cte_sources: Tuple[SqlCteNode, ...]
    join_descs: Tuple[SqlJoinDescription, ...]
    group_bys: Tuple[SqlSelectColumn, ...]
    order_bys: Tuple[SqlOrderByDescription, ...]
    where: Optional[SqlExpressionNode]
    limit: Optional[int]
    distinct: bool

    @staticmethod
    def create(  # noqa: D102
        description: str,
        select_columns: Tuple[SqlSelectColumn, ...],
        from_source: SqlPlanNode,
        from_source_alias: str,
        cte_sources: Tuple[SqlCteNode, ...] = (),
        join_descs: Tuple[SqlJoinDescription, ...] = (),
        group_bys: Tuple[SqlSelectColumn, ...] = (),
        order_bys: Tuple[SqlOrderByDescription, ...] = (),
        where: Optional[SqlExpressionNode] = None,
        limit: Optional[int] = None,
        distinct: bool = False,
    ) -> SqlSelectStatementNode:
        parent_nodes = (from_source,) + tuple(x.right_source for x in join_descs) + cte_sources
        return SqlSelectStatementNode(
            parent_nodes=parent_nodes,
            _description=description,
            select_columns=select_columns,
            from_source=from_source,
            from_source_alias=from_source_alias,
            cte_sources=cte_sources,
            join_descs=join_descs,
            group_bys=group_bys,
            order_bys=order_bys,
            where=where,
            limit=limit,
            distinct=distinct,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_PLAN_SELECT_STATEMENT_ID_PREFIX

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return (
            tuple(super().displayed_properties)
            + tuple(DisplayedProperty(f"col{i}", column) for i, column in enumerate(self.select_columns))
            + (DisplayedProperty("from_source", self.from_source),)
            + tuple(DisplayedProperty(f"join_{i}", join_desc) for i, join_desc in enumerate(self.join_descs))
            + tuple(DisplayedProperty(f"group_by{i}", group_by) for i, group_by in enumerate(self.group_bys))
            + (DisplayedProperty("where", self.where),)
            + tuple(DisplayedProperty(f"order_by{i}", order_by) for i, order_by in enumerate(self.order_bys))
            + (DisplayedProperty("distinct", self.distinct),)
        )

    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_select_statement_node(self)

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D102
        return self

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @property
    @override
    def description(self) -> str:
        return self._description

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return self.select_columns

    @override
    def copy(self) -> SqlSelectStatementNode:
        return SqlSelectStatementNode(
            parent_nodes=self.parent_nodes,
            _description=self._description,
            select_columns=self.select_columns,
            from_source=self.from_source,
            from_source_alias=self.from_source_alias,
            cte_sources=self.cte_sources,
            join_descs=self.join_descs,
            group_bys=self.group_bys,
            order_bys=self.order_bys,
            where=self.where,
            limit=self.limit,
            distinct=self.distinct,
        )


@dataclass(frozen=True, eq=False)
class SqlTableNode(SqlPlanNode):
    """An SQL table that can go in the FROM clause or the JOIN clause."""

    sql_table: SqlTable

    @staticmethod
    def create(sql_table: SqlTable) -> SqlTableNode:  # noqa: D102
        return SqlTableNode(
            parent_nodes=(),
            sql_table=sql_table,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D102
        return f"Read from {self.sql_table.sql}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("table_id", self.sql_table.sql),)

    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_table_node(self)

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D102
        return None

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        if self.sql_table.schema_name is None:
            cte_node = cte_source_mapping.get_cte_node_for_alias(self.sql_table.table_name)
            if cte_node is not None:
                return cte_node.nearest_select_columns(cte_source_mapping)
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return self

    @override
    def copy(self) -> SqlTableNode:
        return SqlTableNode(
            parent_nodes=self.parent_nodes,
            sql_table=self.sql_table,
        )


@dataclass(frozen=True, eq=False)
class SqlSelectQueryFromClauseNode(SqlPlanNode):
    """An SQL select query that can go in the FROM clause.

    Attributes:
        select_query: The SQL select query to include in the FROM clause.
    """

    select_query: str

    @staticmethod
    def create(select_query: str) -> SqlSelectQueryFromClauseNode:  # noqa: D102
        return SqlSelectQueryFromClauseNode(
            parent_nodes=(),
            select_query=select_query,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_PLAN_QUERY_FROM_CLAUSE_ID_PREFIX

    @property
    def description(self) -> str:  # noqa: D102
        return "Read From a Select Query"

    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_query_from_clause_node(self)

    @property
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:  # noqa: D102
        return None

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @override
    def copy(self) -> SqlSelectQueryFromClauseNode:
        return SqlSelectQueryFromClauseNode(parent_nodes=self.parent_nodes, select_query=self.select_query)


@dataclass(frozen=True, eq=False)
class SqlCreateTableAsNode(SqlPlanNode):
    """An SQL node representing a CREATE TABLE AS statement.

    Attributes:
        sql_table: The SQL table to create.
    """

    sql_table: SqlTable

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(sql_table: SqlTable, parent_node: SqlPlanNode) -> SqlCreateTableAsNode:  # noqa: D102
        return SqlCreateTableAsNode(
            parent_nodes=(parent_node,),
            sql_table=sql_table,
        )

    @override
    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_create_table_as_node(self)

    @property
    @override
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @property
    @override
    def description(self) -> str:
        return f"Create table {repr(self.sql_table.sql)}"

    @property
    def parent_node(self) -> SqlPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.SQL_PLAN_CREATE_TABLE_AS_ID_PREFIX

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return self.parent_node.nearest_select_columns(cte_source_mapping)

    @override
    def copy(self) -> SqlCreateTableAsNode:
        return SqlCreateTableAsNode(parent_nodes=self.parent_nodes, sql_table=self.sql_table)


class SqlPlan(MetricFlowDag[SqlPlanNode]):
    """Model for an SQL statement as a DAG."""

    def __init__(self, render_node: SqlPlanNode, plan_id: Optional[DagId] = None) -> None:
        """initializer.

        Args:
            render_node: The node from which to start rendering the SQL statement.
            plan_id: If specified, use this sql_query_plan_id instead of a generated one.
        """
        self._render_node = render_node
        super().__init__(
            dag_id=plan_id or DagId.from_id_prefix(StaticIdPrefix.SQL_PLAN_PREFIX),
            sink_nodes=[self._render_node],
        )

    @property
    def render_node(self) -> SqlPlanNode:  # noqa: D102
        return self._render_node


@dataclass(frozen=True, eq=False)
class SqlCteNode(SqlPlanNode):
    """Represents a single common table expression."""

    select_statement: SqlPlanNode
    cte_alias: str

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(select_statement: SqlPlanNode, cte_alias: str) -> SqlCteNode:  # noqa: D102
        return SqlCteNode(
            parent_nodes=(select_statement,),
            select_statement=select_statement,
            cte_alias=cte_alias,
        )

    def with_new_select(self, new_select_statement: SqlPlanNode) -> SqlCteNode:
        """Return a node with the same attributes but with the new SELECT statement."""
        return SqlCteNode.create(
            select_statement=new_select_statement,
            cte_alias=self.cte_alias,
        )

    @override
    def accept(self, visitor: SqlPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_cte_node(self)

    @property
    @override
    def as_select_node(self) -> Optional[SqlSelectStatementNode]:
        return None

    @property
    @override
    def as_sql_table_node(self) -> Optional[SqlTableNode]:
        return None

    @property
    @override
    def description(self) -> str:
        return "CTE"

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.SQL_PLAN_COMMON_TABLE_EXPRESSION_ID_PREFIX

    @override
    def nearest_select_columns(self, cte_source_mapping: SqlCteAliasMapping) -> Optional[Sequence[SqlSelectColumn]]:
        return self.select_statement.nearest_select_columns(cte_source_mapping)

    @override
    def copy(self) -> SqlCteNode:
        return SqlCteNode(
            parent_nodes=self.parent_nodes,
            select_statement=self.select_statement,
            cte_alias=self.cte_alias,
        )


@dataclass(frozen=True)
class SqlCteAliasMapping(Mergeable):
    """Thin, dict-like object that maps an alias to the associated `SqlCteNode`.

    When merged, the entries from the right mapping take precedence over the entries from the left.
    """

    cte_alias_to_cte_node_items: Tuple[Tuple[str, SqlCteNode], ...] = ()

    @staticmethod
    def create(cte_alias_to_cte_node_mapping: Mapping[str, SqlCteNode]) -> SqlCteAliasMapping:  # noqa: D102
        cte_alias_to_cte_node_pairs = []
        for cte_alias, cte_node in cte_alias_to_cte_node_mapping.items():
            cte_alias_to_cte_node_pairs.append((cte_alias, cte_node))

        return SqlCteAliasMapping(cte_alias_to_cte_node_items=tuple(cte_alias_to_cte_node_pairs))

    @cached_property
    def _cte_alias_to_cte_node_dict(self) -> Mapping[str, SqlCteNode]:
        return {item[0]: item[1] for item in self.cte_alias_to_cte_node_items}

    def get_cte_node_for_alias(self, cte_alias: str) -> Optional[SqlCteNode]:
        """Return the associated `SqlCteNode` for the given alias, or None if the given alias is not known."""
        return self._cte_alias_to_cte_node_dict.get(cte_alias)

    @override
    def merge(self, other: SqlCteAliasMapping) -> SqlCteAliasMapping:
        new_mapping = dict(self._cte_alias_to_cte_node_dict)
        for cte_alias, cte_node in other.cte_alias_to_cte_node_items:
            new_mapping[cte_alias] = cte_node
        return SqlCteAliasMapping.create(new_mapping)

    @classmethod
    @override
    def empty_instance(cls) -> SqlCteAliasMapping:
        return SqlCteAliasMapping()



from __future__ import annotations

from typing import Optional, Sequence, Tuple

from dbt_semantic_interfaces.call_parameter_sets import (
    TimeDimensionCallParameterSet,
)
from dbt_semantic_interfaces.protocols.protocol_hint import ProtocolHint
from dbt_semantic_interfaces.protocols.query_interface import (
    QueryInterfaceTimeDimensionFactory,
)
from dbt_semantic_interfaces.references import EntityReference, TimeDimensionReference
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.errors.error_classes import InvalidQuerySyntax
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
    ResolvedSpecLookUpKey,
)
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.rendered_spec_tracker import RenderedSpecTracker
from metricflow_semantics.specs.where_filter.where_filter_dimension import WhereFilterDimension


class WhereFilterTimeDimension(WhereFilterDimension):
    """A time dimension that is passed in through the where filter parameter."""

    def __str__(self) -> str:
        """Returns the column name.

        Important in the Jinja sandbox.
        """
        call_parameter_set = TimeDimensionCallParameterSet(
            entity_path=self._entity_links,
            time_dimension_reference=TimeDimensionReference(self._element_name),
            time_granularity_name=self._time_granularity_name,
            date_part=self._date_part,
        )

        resolved_spec_key = ResolvedSpecLookUpKey(
            filter_location=self._where_filter_location,
            call_parameter_set=call_parameter_set,
        )

        resolved_spec = self._resolved_spec_lookup.checked_resolved_spec(resolved_spec_key)
        resolved_elements = self._resolved_spec_lookup.checked_resolved_linkable_elements(resolved_spec_key)
        self._rendered_spec_tracker.record_rendered_spec_to_elements_mapping((resolved_spec, resolved_elements))
        column_association = self._column_association_resolver.resolve_spec(resolved_spec)

        return column_association.column_name


class WhereFilterTimeDimensionFactory(ProtocolHint[QueryInterfaceTimeDimensionFactory]):
    """Creates a WhereFilterTimeDimension.

    Each call to `create` adds a TimeDimensionSpec to time_dimension_specs.
    """

    @override
    def _implements_protocol(self) -> QueryInterfaceTimeDimensionFactory:
        return self

    def __init__(  # noqa
        self,
        column_association_resolver: ColumnAssociationResolver,
        spec_resolution_lookup: FilterSpecResolutionLookUp,
        where_filter_location: WhereFilterLocation,
        rendered_spec_tracker: RenderedSpecTracker,
        custom_granularity_names: Tuple[str, ...],
    ):
        self._column_association_resolver = column_association_resolver
        self._resolved_spec_lookup = spec_resolution_lookup
        self._where_filter_location = where_filter_location
        self._rendered_spec_tracker = rendered_spec_tracker
        self._custom_granularity_names = custom_granularity_names

    def create(
        self,
        time_dimension_name: str,
        time_granularity_name: Optional[str] = None,
        entity_path: Sequence[str] = (),
        descending: Optional[bool] = None,
        date_part_name: Optional[str] = None,
    ) -> WhereFilterTimeDimension:
        """Create a WhereFilterTimeDimension."""
        if descending:
            raise InvalidQuerySyntax(
                "Can't set descending in the where clause. Try setting descending in the order_by clause instead"
            )

        time_granularity_name = time_granularity_name.lower() if time_granularity_name else None
        structured_name = StructuredLinkableSpecName.from_name(
            qualified_name=time_dimension_name.lower(), custom_granularity_names=self._custom_granularity_names
        )

        if (
            structured_name.time_granularity_name
            and time_granularity_name
            and structured_name.time_granularity_name != time_granularity_name
        ):
            raise InvalidQuerySyntax(
                f"Received different granularities in `time_dimension_name` parameter ('{time_dimension_name}') "
                f"and `time_granularity_name` parameter ('{time_granularity_name}')."
            )

        return WhereFilterTimeDimension(
            column_association_resolver=self._column_association_resolver,
            resolved_spec_lookup=self._resolved_spec_lookup,
            where_filter_location=self._where_filter_location,
            rendered_spec_tracker=self._rendered_spec_tracker,
            element_name=structured_name.element_name,
            entity_links=tuple(
                EntityReference(entity_link_name.lower())
                for entity_link_name in (tuple(entity_path) + structured_name.entity_link_names)
            ),
            time_granularity_name=time_granularity_name or structured_name.time_granularity_name,
            date_part=DatePart(date_part_name.lower()) if date_part_name else None,
        )



from __future__ import annotations

from typing import Optional, Sequence

from dbt_semantic_interfaces.call_parameter_sets import (
    EntityCallParameterSet,
)
from dbt_semantic_interfaces.naming.dundered import DunderedNameFormatter
from dbt_semantic_interfaces.protocols.protocol_hint import ProtocolHint
from dbt_semantic_interfaces.protocols.query_interface import QueryInterfaceEntity, QueryInterfaceEntityFactory
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.type_enums import TimeGranularity
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.errors.error_classes import InvalidQuerySyntax
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
    ResolvedSpecLookUpKey,
)
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.rendered_spec_tracker import RenderedSpecTracker


class WhereFilterEntity(ProtocolHint[QueryInterfaceEntity]):
    """An entity that is passed in through the where filter parameter."""

    @override
    def _implements_protocol(self) -> QueryInterfaceEntity:
        return self

    def __init__(  # noqa
        self,
        column_association_resolver: ColumnAssociationResolver,
        resolved_spec_lookup: FilterSpecResolutionLookUp,
        where_filter_location: WhereFilterLocation,
        rendered_spec_tracker: RenderedSpecTracker,
        element_name: str,
        entity_links: Sequence[EntityReference],
        time_grain: Optional[TimeGranularity] = None,
        date_part: Optional[DatePart] = None,
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._resolved_spec_lookup = resolved_spec_lookup
        self._where_filter_location = where_filter_location
        self._rendered_spec_tracker = rendered_spec_tracker
        self._element_name = element_name
        self._entity_links = tuple(entity_links)
        self._time_grain = time_grain
        self._date_part = date_part

    def descending(self, _is_descending: bool) -> QueryInterfaceEntity:
        """Set the sort order for order-by."""
        raise InvalidQuerySyntax(
            "Can't set descending in the where clause. Try setting descending in the order_by clause instead"
        )

    def __str__(self) -> str:
        """Returns the column name.

        Important in the Jinja sandbox.
        """
        call_parameter_set = EntityCallParameterSet(
            entity_path=self._entity_links,
            entity_reference=EntityReference(self._element_name),
        )

        resolved_spec_key = ResolvedSpecLookUpKey(
            filter_location=self._where_filter_location,
            call_parameter_set=call_parameter_set,
        )

        resolved_spec = self._resolved_spec_lookup.checked_resolved_spec(resolved_spec_key)
        resolved_elements = self._resolved_spec_lookup.checked_resolved_linkable_elements(resolved_spec_key)
        self._rendered_spec_tracker.record_rendered_spec_to_elements_mapping((resolved_spec, resolved_elements))
        column_association = self._column_association_resolver.resolve_spec(resolved_spec)

        return column_association.column_name


class WhereFilterEntityFactory(ProtocolHint[QueryInterfaceEntityFactory]):
    """Creates a WhereFilterEntity.

    Each call to `create` adds an EntitySpec to entity_specs.
    """

    @override
    def _implements_protocol(self) -> QueryInterfaceEntityFactory:
        return self

    def __init__(  # noqa
        self,
        column_association_resolver: ColumnAssociationResolver,
        spec_resolution_lookup: FilterSpecResolutionLookUp,
        where_filter_location: WhereFilterLocation,
        rendered_spec_tracker: RenderedSpecTracker,
    ):
        self._column_association_resolver = column_association_resolver
        self._resolved_spec_lookup = spec_resolution_lookup
        self._where_filter_location = where_filter_location
        self._rendered_spec_tracker = rendered_spec_tracker

    def create(self, entity_name: str, entity_path: Sequence[str] = ()) -> WhereFilterEntity:
        """Create a WhereFilterEntity."""
        structured_name = DunderedNameFormatter.parse_name(entity_name.lower())

        return WhereFilterEntity(
            column_association_resolver=self._column_association_resolver,
            resolved_spec_lookup=self._resolved_spec_lookup,
            where_filter_location=self._where_filter_location,
            rendered_spec_tracker=self._rendered_spec_tracker,
            element_name=structured_name.element_name,
            entity_links=tuple(EntityReference(entity_link_name.lower()) for entity_link_name in entity_path)
            + structured_name.entity_links,
        )



from __future__ import annotations

import logging
from typing import Optional, Sequence

from dbt_semantic_interfaces.call_parameter_sets import (
    DimensionCallParameterSet,
    EntityCallParameterSet,
    MetricCallParameterSet,
    TimeDimensionCallParameterSet,
)
from dbt_semantic_interfaces.naming.keywords import DUNDER
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet, InstanceSpecSetTransform, group_spec_by_type

logger = logging.getLogger(__name__)


class ObjectBuilderNameConverter:
    """Methods for converting classes related to the object-builder naming scheme to strings.

    String conversions are necessary for parsing user input and generating error messages.

    TODO: CallParameterSets / QueryParameter objects need restructuring.
    """

    @staticmethod
    def input_str_from_entity_call_parameter_set(parameter_set: EntityCallParameterSet) -> str:  # noqa: D102
        initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
            element_name=parameter_set.entity_reference.element_name,
            entity_links=parameter_set.entity_path,
            group_by=(),
            time_granularity_name=None,
            date_part=None,
        )
        return f"Entity({initializer_parameter_str})"

    @staticmethod
    def input_str_from_metric_call_parameter_set(parameter_set: MetricCallParameterSet) -> str:  # noqa: D102
        initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
            element_name=parameter_set.metric_reference.element_name,
            entity_links=(),
            group_by=tuple(
                EntityReference(element_name=group_by_ref.element_name) for group_by_ref in parameter_set.group_by
            ),
            time_granularity_name=None,
            date_part=None,
        )
        return f"Metric({initializer_parameter_str})"

    @staticmethod
    def initializer_parameter_str(
        element_name: str,
        entity_links: Sequence[EntityReference],
        group_by: Sequence[EntityReference],
        time_granularity_name: Optional[str],
        date_part: Optional[DatePart],
    ) -> str:
        """Return the parameters that should go in the initializer.

        e.g. `'user__country', time_granularity_name='month'`
        """
        initializer_parameters = []
        entity_link_names = list(entity_link.element_name for entity_link in entity_links)
        if len(entity_link_names) > 0:
            initializer_parameters.append(repr(entity_link_names[-1] + DUNDER + element_name))
        else:
            initializer_parameters.append(repr(element_name))
        if time_granularity_name is not None:
            initializer_parameters.append(
                f"'{time_granularity_name}'",
            )
        if date_part is not None:
            initializer_parameters.append(f"date_part_name={repr(date_part.value)}")
        if len(entity_link_names) > 1:
            initializer_parameters.append(f"entity_path={repr(entity_link_names[:-1])}")
        if group_by:
            initializer_parameters.append(f"group_by={[group_by_ref.element_name for group_by_ref in group_by]}")

        return ", ".join(initializer_parameters)

    class _ObjectBuilderNameTransform(InstanceSpecSetTransform[Sequence[str]]):
        """Transforms specs into strings following the object builder scheme."""

        @override
        def transform(self, spec_set: InstanceSpecSet) -> Sequence[str]:
            names_to_return = []

            for entity_spec in spec_set.entity_specs:
                initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
                    element_name=entity_spec.element_name,
                    entity_links=entity_spec.entity_links,
                    group_by=(),
                    time_granularity_name=None,
                    date_part=None,
                )
                names_to_return.append(f"Entity({initializer_parameter_str})")

            for dimension_spec in spec_set.dimension_specs:
                initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
                    element_name=dimension_spec.element_name,
                    entity_links=dimension_spec.entity_links,
                    group_by=(),
                    time_granularity_name=None,
                    date_part=None,
                )
                names_to_return.append(f"Dimension({initializer_parameter_str})")

            for time_dimension_spec in spec_set.time_dimension_specs:
                initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
                    element_name=time_dimension_spec.element_name,
                    entity_links=time_dimension_spec.entity_links,
                    group_by=(),
                    time_granularity_name=time_dimension_spec.time_granularity.name,
                    date_part=time_dimension_spec.date_part,
                )
                names_to_return.append(f"TimeDimension({initializer_parameter_str})")

            for group_by_metric_spec in spec_set.group_by_metric_specs:
                initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
                    element_name=group_by_metric_spec.element_name,
                    entity_links=(),
                    group_by=group_by_metric_spec.entity_links,
                    time_granularity_name=None,
                    date_part=None,
                )
                names_to_return.append(f"Metric({initializer_parameter_str})")

            return names_to_return

    @staticmethod
    def input_str_from_spec(instance_spec: InstanceSpec) -> Optional[str]:  # noqa: D102
        names = ObjectBuilderNameConverter._ObjectBuilderNameTransform().transform(group_spec_by_type(instance_spec))

        if len(names) == 0:
            return None
        elif len(names) > 1:
            raise RuntimeError(str(LazyFormat("Expected at most one name", instance_spec=instance_spec, names=names)))

        return names[0]

    @staticmethod
    def input_str_from_dimension_call_parameter_set(parameter_set: DimensionCallParameterSet) -> str:  # noqa: D102
        initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
            element_name=parameter_set.dimension_reference.element_name,
            entity_links=parameter_set.entity_path,
            group_by=(),
            time_granularity_name=None,
            date_part=None,
        )
        return f"Dimension({initializer_parameter_str})"

    @staticmethod
    def input_str_from_time_dimension_call_parameter_set(  # noqa: D102
        parameter_set: TimeDimensionCallParameterSet,
    ) -> str:
        initializer_parameter_str = ObjectBuilderNameConverter.initializer_parameter_str(
            element_name=parameter_set.time_dimension_reference.element_name,
            entity_links=parameter_set.entity_path,
            group_by=(),
            time_granularity_name=None,
            date_part=None,
        )
        return f"TimeDimension({initializer_parameter_str})"



from __future__ import annotations

import collections
from typing import Dict, Generic, List, TypeVar

ElementX = TypeVar("ElementX")
ElementY = TypeVar("ElementY")


class ElementGrouper(Generic[ElementX, ElementY]):
    """Groups an Element pair such that we have (ElementX, List[ElementY])."""

    def __init__(self) -> None:  # noqa: D107
        self._groups: Dict[ElementX, List[ElementY]] = collections.defaultdict(list)

    def add_value(self, key: ElementX, value: ElementY) -> None:  # noqa: D102
        self._groups[key].append(value)

    def get_values(self, key: ElementX) -> List[ElementY]:  # noqa: D102
        if key not in self._groups:
            raise KeyError(f"Unable to find `{key}` in ElementGrouper")
        return self._groups[key]

    @property
    def keys(self) -> List[ElementX]:  # noqa :D
        return list(self._groups.keys())



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Dict, FrozenSet, Iterable, Set

from metricflow.sql.sql_plan import (
    SqlPlanNode,
)

logger = logging.getLogger(__name__)


class NodeToColumnAliasMapping:
    """Mutable class for mapping a SQL node to an arbitrary set of column aliases for that node.

    * Alternatively, this can be described as mapping a location in the SQL query plan to a set of column aliases.
    * See `SqlMapRequiredColumnAliasesVisitor` for the main use case for this class.
    * This is a thin wrapper over a dict to aid readability.
    """

    def __init__(self) -> None:  # noqa: D107
        self._node_to_tagged_aliases: Dict[SqlPlanNode, Set[str]] = defaultdict(set)

    def get_aliases(self, node: SqlPlanNode) -> FrozenSet[str]:
        """Return the column aliases added for the given SQL node."""
        return frozenset(self._node_to_tagged_aliases[node])

    def add_alias(self, node: SqlPlanNode, column_alias: str) -> None:  # noqa: D102
        return self._node_to_tagged_aliases[node].add(column_alias)

    def add_aliases(self, node: SqlPlanNode, column_aliases: Iterable[str]) -> None:  # noqa: D102
        self._node_to_tagged_aliases[node].update(column_aliases)



from __future__ import annotations

import logging
from typing import Optional, Sequence, Union

from dbt_semantic_interfaces.call_parameter_sets import (
    DimensionCallParameterSet,
    TimeDimensionCallParameterSet,
)
from dbt_semantic_interfaces.naming.dundered import DunderedNameFormatter
from dbt_semantic_interfaces.protocols.protocol_hint import ProtocolHint
from dbt_semantic_interfaces.protocols.query_interface import (
    QueryInterfaceDimension,
    QueryInterfaceDimensionFactory,
)
from dbt_semantic_interfaces.references import DimensionReference, EntityReference, TimeDimensionReference
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.errors.error_classes import InvalidQuerySyntax
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
    ResolvedSpecLookUpKey,
)
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.rendered_spec_tracker import RenderedSpecTracker

logger = logging.getLogger(__name__)


class WhereFilterDimension(ProtocolHint[QueryInterfaceDimension]):
    """A dimension that is passed in through the where filter parameter."""

    @override
    def _implements_protocol(self) -> QueryInterfaceDimension:
        return self

    def __init__(  # noqa
        self,
        column_association_resolver: ColumnAssociationResolver,
        resolved_spec_lookup: FilterSpecResolutionLookUp,
        where_filter_location: WhereFilterLocation,
        rendered_spec_tracker: RenderedSpecTracker,
        element_name: str,
        entity_links: Sequence[EntityReference],
        time_granularity_name: Optional[str] = None,
        date_part: Optional[DatePart] = None,
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._resolved_spec_lookup = resolved_spec_lookup
        self._where_filter_location = where_filter_location
        self._rendered_spec_tracker = rendered_spec_tracker
        self._element_name = element_name
        self._entity_links = tuple(entity_links)
        self._time_granularity_name = time_granularity_name
        self._date_part = date_part

    def grain(self, time_granularity_name: str) -> QueryInterfaceDimension:
        """The time granularity."""
        return WhereFilterDimension(
            column_association_resolver=self._column_association_resolver,
            resolved_spec_lookup=self._resolved_spec_lookup,
            where_filter_location=self._where_filter_location,
            rendered_spec_tracker=self._rendered_spec_tracker,
            element_name=self._element_name,
            entity_links=self._entity_links,
            time_granularity_name=time_granularity_name.lower(),
            date_part=self._date_part,
        )

    def date_part(self, date_part_name: str) -> QueryInterfaceDimension:
        """The date_part requested to extract."""
        return WhereFilterDimension(
            column_association_resolver=self._column_association_resolver,
            resolved_spec_lookup=self._resolved_spec_lookup,
            where_filter_location=self._where_filter_location,
            rendered_spec_tracker=self._rendered_spec_tracker,
            element_name=self._element_name,
            entity_links=self._entity_links,
            time_granularity_name=self._time_granularity_name,
            date_part=DatePart(date_part_name.lower()),
        )

    def descending(self, _is_descending: bool) -> QueryInterfaceDimension:
        """Set the sort order for order-by."""
        raise InvalidQuerySyntax("descending is invalid in the where parameter")

    def __str__(self) -> str:
        """Returns the column name.

        Important in the Jinja sandbox.
        """
        call_parameter_set: Union[TimeDimensionCallParameterSet, DimensionCallParameterSet]
        if self._time_granularity_name is not None or self._date_part is not None:
            call_parameter_set = TimeDimensionCallParameterSet(
                entity_path=self._entity_links,
                time_dimension_reference=TimeDimensionReference(self._element_name),
                time_granularity_name=self._time_granularity_name,
                date_part=self._date_part,
            )
        else:
            call_parameter_set = DimensionCallParameterSet(
                entity_path=self._entity_links,
                dimension_reference=DimensionReference(self._element_name),
            )

        resolved_spec_key = ResolvedSpecLookUpKey(
            filter_location=self._where_filter_location,
            call_parameter_set=call_parameter_set,
        )

        resolved_spec = self._resolved_spec_lookup.checked_resolved_spec(resolved_spec_key)
        resolved_elements = self._resolved_spec_lookup.checked_resolved_linkable_elements(resolved_spec_key)
        self._rendered_spec_tracker.record_rendered_spec_to_elements_mapping((resolved_spec, resolved_elements))
        column_association = self._column_association_resolver.resolve_spec(resolved_spec)

        return column_association.column_name


class WhereFilterDimensionFactory(ProtocolHint[QueryInterfaceDimensionFactory]):
    """Creates a WhereFilterDimension.

    Each call to `create` adds a WhereFilterDimension to created.
    """

    @override
    def _implements_protocol(self) -> QueryInterfaceDimensionFactory:
        return self

    def __init__(  # noqa
        self,
        column_association_resolver: ColumnAssociationResolver,
        spec_resolution_lookup: FilterSpecResolutionLookUp,
        where_filter_location: WhereFilterLocation,
        rendered_spec_tracker: RenderedSpecTracker,
    ):
        self._column_association_resolver = column_association_resolver
        self._resolved_spec_lookup = spec_resolution_lookup
        self._where_filter_location = where_filter_location
        self._rendered_spec_tracker = rendered_spec_tracker

    def create(self, name: str, entity_path: Sequence[str] = ()) -> WhereFilterDimension:
        """Create a WhereFilterDimension."""
        structured_name = DunderedNameFormatter.parse_name(name.lower())

        return WhereFilterDimension(
            column_association_resolver=self._column_association_resolver,
            resolved_spec_lookup=self._resolved_spec_lookup,
            where_filter_location=self._where_filter_location,
            rendered_spec_tracker=self._rendered_spec_tracker,
            element_name=structured_name.element_name,
            entity_links=tuple(EntityReference(entity_link_name.lower()) for entity_link_name in entity_path)
            + structured_name.entity_links,
        )



"""Utility module for building sql expressions from inputs derived from dataflow plan or other nodes."""

from __future__ import annotations

from typing import List, Sequence

from metricflow_semantics.sql.sql_exprs import (
    SqlAggregateFunctionExpression,
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlExpressionNode,
    SqlFunction,
)


def make_coalesced_expr(table_aliases: Sequence[str], column_alias: str) -> SqlExpressionNode:
    """Makes a coalesced expression of the given column from the given table aliases.

    e.g.

    table_aliases = ["a", "b"]
    column_alias = "is_instant"

    ->

    COALESCE(a.is_instant, b.is_instant)
    """
    if len(table_aliases) == 1:
        return SqlColumnReferenceExpression.create(
            col_ref=SqlColumnReference(
                table_alias=table_aliases[0],
                column_name=column_alias,
            )
        )
    else:
        columns_to_coalesce: List[SqlExpressionNode] = []
        for table_alias in table_aliases:
            columns_to_coalesce.append(
                SqlColumnReferenceExpression.create(
                    col_ref=SqlColumnReference(
                        table_alias=table_alias,
                        column_name=column_alias,
                    )
                )
            )
        return SqlAggregateFunctionExpression.create(
            sql_function=SqlFunction.COALESCE,
            sql_function_args=columns_to_coalesce,
        )



"""Determines whether we can compute a valid join for the given inputs.

For example:

I have a node containing a measure that's needed for a metric, and I also have a list of dimensions that I need for that
measure.

Can I join that node with other nodes using a common entity to get those dimensions? If so, what are those nodes,
what entity should I join on, and if I do that join, what dimensions can be retrieved?

Note: the term "dimension" is used below, but it actually refers to any LinkableInstance. Also, when you see the term
"left_node", think "node containing the measure". Using the term "left_node" as the scenario of joining measure nodes
to dimension nodes is an easy case to explain, but there are non-measure nodes that can be joined with dimension nodes
to realize other planned features.
"""

from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from metricflow_semantics.instances import InstanceSet
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import SemanticModelJoinEvaluator
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.specs.entity_spec import LinklessEntitySpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.sql.sql_join_type import SqlJoinType

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.partitions import (
    PartitionDimensionJoinDescription,
    PartitionJoinResolver,
    PartitionTimeDimensionJoinDescription,
)
from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_to_base import JoinDescription, ValidityWindowJoinDescription
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataset.dataset_classes import DataSet
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.plan_conversion.instance_converters import CreateValidityWindowJoinDescription

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class JoinLinkableInstancesRecipe:
    """A recipe for how to join a node containing linkable instances to the "left_node".

    node_to_join contains the linkable instances that are needed - it should be filtered so that the output data set of
    that node only includes the entity instance for the join and the instances associated with
    satisfiable_linkable_specs.
    """

    node_to_join: DataflowPlanNode
    # The entity to join "node_to_join" on. Only nullable if using CROSS JOIN.
    join_on_entity: Optional[LinklessEntitySpec]
    # The linkable instances from the query that can be satisfied if we join this node. Note that this is different from
    # the linkable specs in the node that can help to satisfy the query. e.g. "user_id__country" might be one of the
    # "satisfiable_linkable_specs", but "country" is the linkable spec in the node.
    satisfiable_linkable_specs: List[LinkableInstanceSpec]
    # Join type to use when joining nodes
    join_type: SqlJoinType

    # The partitions to join on, if there are matching partitions between the left_node and node_to_join.
    join_on_partition_dimensions: Tuple[PartitionDimensionJoinDescription, ...]
    join_on_partition_time_dimensions: Tuple[PartitionTimeDimensionJoinDescription, ...]

    validity_window: Optional[ValidityWindowJoinDescription] = None

    def __post_init__(self) -> None:  # noqa: D105
        if self.join_on_entity is None and self.join_type != SqlJoinType.CROSS_JOIN:
            raise RuntimeError("`join_on_entity` is required unless using CROSS JOIN.")

    # TODO: JoinDescription is very similar to JoinLinkableInstancesRecipe. Can we consolidate by just adding a
    # `filtered_node_to_join` property on JoinLinkableInstancesRecipe?
    @property
    def join_description(self) -> JoinDescription:
        """The recipe as a join description to use in the dataflow plan node.

        Here, we figure out which instance specs to keep from this node in order to join to it and render its
        satisfiable linkable specs, e.g. if the node is used to satisfy "user_id__country", the node must have the
        entity "user_id" and the "country" dimension so that it can be joined to the source node.
        """
        include_specs: List[LinkableInstanceSpec] = []
        assert all(
            [
                len(spec.entity_links) > 0
                for spec in self.satisfiable_linkable_specs
                if not spec.element_name == METRIC_TIME_ELEMENT_NAME
            ]
        )

        # Get the specs needed to join onto this node.
        if self.node_to_join.aggregated_to_elements:
            include_specs.extend(self.node_to_join.aggregated_to_elements)
        else:
            for spec in self.satisfiable_linkable_specs:
                if len(spec.entity_links) > 0:
                    include_specs.append(LinklessEntitySpec.from_reference(spec.entity_links[0]))

        include_specs.extend([join.node_to_join_dimension_spec for join in self.join_on_partition_dimensions])
        include_specs.extend([join.node_to_join_time_dimension_spec for join in self.join_on_partition_time_dimensions])
        if self.validity_window:
            include_specs.extend(
                [self.validity_window.window_start_dimension, self.validity_window.window_end_dimension]
            )

        # `satisfiable_linkable_specs` describes what can be satisfied after the join, so remove the entity
        # link when filtering before the join.
        include_specs.extend(
            [
                spec.without_first_entity_link if len(spec.entity_links) > 0 else spec
                for spec in self.satisfiable_linkable_specs
            ]
        )

        filtered_node_to_join = FilterElementsNode.create(
            parent_node=self.node_to_join, include_specs=group_specs_by_type(include_specs).dedupe()
        )

        return JoinDescription(
            join_node=filtered_node_to_join,
            join_on_entity=self.join_on_entity,
            join_on_partition_dimensions=self.join_on_partition_dimensions,
            join_on_partition_time_dimensions=self.join_on_partition_time_dimensions,
            validity_window=self.validity_window,
            join_type=self.join_type,
        )


@dataclass(frozen=True)
class LinkableInstanceSatisfiabilityEvaluation:
    """Evaluation for a node on the ability to get linkable instances that we need to resolve a query.

    Includes whether the node contains the linkable instances that we need (local_linkable_specs), or if it can have
    them joined in (joinable_linkable_specs). Linkable instances that can't be satisfied either locally or via joins are
    listed in unjoinable_linkable_specs.
    """

    local_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    joinable_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    join_recipes: Tuple[JoinLinkableInstancesRecipe, ...]
    unjoinable_linkable_specs: Tuple[LinkableInstanceSpec, ...]


class NodeEvaluatorForLinkableInstances:
    """Helps to evaluate if linkable instances can be obtained using the given node, with joins if necessary.

    For example, consider a "left_node" (that will be on the left side of the join) containing the "bookings" measure,
    "is_instant" dimension, and "listing_id" entity with nodes_available_for_joins including a node with the
    "listing_id" entity, and the "country" dimension.

    We want to know if we can get "bookings", "is_instant", "listing_id__country" using the left_node. The result
    should be that we know: "is_instant" is available locally (i.e. in the same node), and if we join another node
    (on the right side of the join) containing "listing_id" and "country" by "listing_id", we can get
    "listing_id__country".

    """

    def __init__(
        self,
        semantic_model_lookup: SemanticModelLookup,
        nodes_available_for_joins: Sequence[DataflowPlanNode],
        node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver,
        time_spine_metric_time_nodes: Sequence[MetricTimeDimensionTransformNode],
    ) -> None:
        """Initializer.

        Args:
            semantic_model_lookup: Needed to resolve partition dimensions.
            nodes_available_for_joins: Nodes that contain linkable instances and may be joined with "left_node" (on the
            left side of the join and containing a desired measure) to retrieve the needed linkable instances.
            node_data_set_resolver: Figures out what data set is output by a node.
            time_spine_node: If nodes_available_for_joins contains a time spine node, it should be identical to this
            one as there is logic to check for equality.
        """
        self._semantic_model_lookup = semantic_model_lookup
        self._nodes_available_for_joins = nodes_available_for_joins
        self._node_data_set_resolver = node_data_set_resolver
        self._partition_resolver = PartitionJoinResolver(self._semantic_model_lookup)
        self._join_evaluator = SemanticModelJoinEvaluator(self._semantic_model_lookup)
        self._time_spine_metric_time_nodes = time_spine_metric_time_nodes

    def _find_joinable_candidate_nodes_that_can_satisfy_linkable_specs(
        self,
        left_node_instance_set: InstanceSet,
        needed_linkable_specs: List[LinkableInstanceSpec],
        default_join_type: SqlJoinType,
    ) -> List[JoinLinkableInstancesRecipe]:
        """Get nodes that can be joined to get 1 or more of the "needed_linkable_specs".

        The returned list is ordered by the number of "needed_linkable_specs" that it can satisfy.
        """
        candidates_for_join: List[JoinLinkableInstancesRecipe] = []
        left_node_spec_set = left_node_instance_set.spec_set
        for right_node in self._nodes_available_for_joins:
            data_set_in_right_node: SqlDataSet = self._node_data_set_resolver.get_output_data_set(right_node)
            linkable_specs_in_right_node = data_set_in_right_node.instance_set.spec_set.linkable_specs

            # If right node is time spine source node, use cross join.
            if right_node in self._time_spine_metric_time_nodes:
                satisfiable_metric_time_specs = [
                    spec for spec in linkable_specs_in_right_node if spec in needed_linkable_specs
                ]
                candidates_for_join.append(
                    JoinLinkableInstancesRecipe(
                        node_to_join=right_node,
                        join_on_entity=None,
                        satisfiable_linkable_specs=list(satisfiable_metric_time_specs),
                        join_on_partition_dimensions=(),
                        join_on_partition_time_dimensions=(),
                        join_type=SqlJoinType.CROSS_JOIN,
                    )
                )
                continue

            entity_specs_in_right_node = data_set_in_right_node.instance_set.spec_set.entity_specs

            # For each unlinked entity in the data set, create a candidate for joining.
            # For a data set to be useful for satisfying a linkable spec, it needs to have the entity
            # and the linkable spec without the entity. This allows joining based on the entity, which will
            # then produce the linkable spec. See comments further below for more details.

            for entity_spec_in_right_node in entity_specs_in_right_node:
                entity_instance_in_right_node = None
                for instance in data_set_in_right_node.instance_set.entity_instances:
                    if instance.spec == entity_spec_in_right_node:
                        entity_instance_in_right_node = instance
                        break

                if entity_instance_in_right_node is None:
                    raise RuntimeError(f"Could not find entity instance with name ({entity_spec_in_right_node})")

                assert (
                    len(entity_instance_in_right_node.defined_from) == 1
                ), f"Did not get exactly 1 defined_from in {entity_instance_in_right_node}"

                entity_in_right_node = self._semantic_model_lookup.get_entity_in_semantic_model(
                    entity_instance_in_right_node.defined_from[0]
                )
                if entity_in_right_node is None:
                    raise RuntimeError(
                        f"Invalid SemanticModelElementReference {entity_instance_in_right_node.defined_from[0]}"
                    )

                entity_instance_in_left_node = None
                for instance in left_node_instance_set.entity_instances:
                    if instance.spec.reference == entity_spec_in_right_node.reference:
                        entity_instance_in_left_node = instance
                        break

                if entity_instance_in_left_node is None:
                    # The right node can have a superset of entities.
                    continue

                assert len(entity_instance_in_left_node.defined_from) == 1
                assert len(entity_instance_in_right_node.defined_from) == 1

                entity_spec_matches_aggregated_specs = {
                    spec.reference for spec in right_node.aggregated_to_elements
                } == {entity_spec_in_right_node.reference}
                if not (
                    self._join_evaluator.is_valid_semantic_model_join(
                        left_semantic_model_reference=entity_instance_in_left_node.defined_from[
                            0
                        ].semantic_model_reference,
                        right_semantic_model_reference=entity_instance_in_right_node.defined_from[
                            0
                        ].semantic_model_reference,
                        on_entity_reference=entity_spec_in_right_node.reference,
                    )
                    or entity_spec_matches_aggregated_specs
                ):
                    continue

                linkless_entity_spec_in_node = LinklessEntitySpec.from_element_name(
                    entity_spec_in_right_node.element_name
                )

                satisfiable_linkable_specs = []
                for needed_linkable_spec in needed_linkable_specs:
                    if len(needed_linkable_spec.entity_links) == 0:
                        assert (
                            needed_linkable_spec.element_name == METRIC_TIME_ELEMENT_NAME
                        ), "Only metric_time should have 0 entity links."
                        continue

                    # If the entity in the data set matches the link, then it can be used for joins. For example,
                    # if the node has the entity "user_id", and dimension "country" then it can be used for
                    # satisfying "user_id__country".
                    #
                    # Multi-hop example:
                    # required_linkable_spec = "user_id__device_id__platform"
                    # entity_spec_in_data_set = "user_id"
                    #
                    # Then the data set must contain "device_id__platform", which is realized with
                    #
                    # required_linkable_spec.remove_first_entity_link()
                    #
                    # We might also need to check the entity type and see if it's the type of join we're allowing,
                    # but since we're doing all left joins now, it's been left out.

                    required_entity_matches_data_set_entity = (
                        LinklessEntitySpec.from_reference(needed_linkable_spec.entity_links[0])
                        == linkless_entity_spec_in_node
                    )
                    needed_linkable_spec_in_node = (
                        needed_linkable_spec.without_first_entity_link in linkable_specs_in_right_node
                    )
                    if required_entity_matches_data_set_entity and needed_linkable_spec_in_node:
                        satisfiable_linkable_specs.append(needed_linkable_spec)

                # If this node can satisfy some linkable specs, it could be useful to join on, so add it to the
                # candidate list.
                if len(satisfiable_linkable_specs) > 0:
                    join_on_partition_dimensions = self._partition_resolver.resolve_partition_dimension_joins(
                        left_node_spec_set=left_node_spec_set,
                        node_to_join_spec_set=data_set_in_right_node.instance_set.spec_set,
                    )
                    join_on_partition_time_dimensions = self._partition_resolver.resolve_partition_time_dimension_joins(
                        left_node_spec_set=left_node_spec_set,
                        node_to_join_spec_set=data_set_in_right_node.instance_set.spec_set,
                    )
                    validity_window_join_description = CreateValidityWindowJoinDescription(
                        self._semantic_model_lookup
                    ).transform(instance_set=data_set_in_right_node.instance_set)

                    candidates_for_join.append(
                        JoinLinkableInstancesRecipe(
                            node_to_join=right_node,
                            join_on_entity=linkless_entity_spec_in_node,
                            satisfiable_linkable_specs=satisfiable_linkable_specs,
                            join_on_partition_dimensions=join_on_partition_dimensions,
                            join_on_partition_time_dimensions=join_on_partition_time_dimensions,
                            validity_window=validity_window_join_description,
                            join_type=default_join_type,
                        )
                    )

        # Return with the candidate set that can satisfy the most linkable specs at the front.
        return sorted(
            candidates_for_join,
            key=lambda x: len(x.satisfiable_linkable_specs),
            reverse=True,
        )

    @staticmethod
    def _update_candidates_that_can_satisfy_linkable_specs(
        candidates_for_join: List[JoinLinkableInstancesRecipe],
        already_satisfisfied_linkable_specs: List[LinkableInstanceSpec],
    ) -> List[JoinLinkableInstancesRecipe]:
        """Update / filter candidates_for_join based on linkable instance specs that we have already satisfied.

        Those linkable instances are no longer needed because those were satisfied previously by joining a node
        containing some needed linkable instances to the "left_node". This method will then remove the satisfied
        linkable instances from the candidate recipes, and if a candidate recipe doesn't have any linkable instances
        that can help satisfy the query, it is removed.
        """
        updated_candidate_data_sets: List[JoinLinkableInstancesRecipe] = []
        for candidate_for_join in candidates_for_join:
            updated_satisfiable_linkable_specs = list(
                set(candidate_for_join.satisfiable_linkable_specs) - set(already_satisfisfied_linkable_specs)
            )

            if len(updated_satisfiable_linkable_specs) > 0:
                updated_candidate_data_sets.append(
                    JoinLinkableInstancesRecipe(
                        node_to_join=candidate_for_join.node_to_join,
                        join_on_entity=candidate_for_join.join_on_entity,
                        satisfiable_linkable_specs=updated_satisfiable_linkable_specs,
                        join_on_partition_dimensions=candidate_for_join.join_on_partition_dimensions,
                        join_on_partition_time_dimensions=candidate_for_join.join_on_partition_time_dimensions,
                        validity_window=candidate_for_join.validity_window,
                        join_type=candidate_for_join.join_type,
                    )
                )
        return sorted(
            updated_candidate_data_sets,
            key=lambda x: len(x.satisfiable_linkable_specs),
            reverse=True,
        )

    def evaluate_node(
        self,
        left_node: DataflowPlanNode,
        required_linkable_specs: Sequence[LinkableInstanceSpec],
        default_join_type: SqlJoinType,
    ) -> LinkableInstanceSatisfiabilityEvaluation:
        """Evaluates if the "required_linkable_specs" can be realized by joining the "left_node" with other nodes.

        In other words, given the data set associated with "left_node":

        * Can all "required_linkable_specs" be retrieved from the left_node? (These would be considered "local").
        * If not, can they be retrieved by joining an available node though a common entity?
        * If so, return all possible ways (by joining different nodes) that can be done.
        """
        candidate_instance_set: InstanceSet = self._node_data_set_resolver.get_output_data_set(left_node).instance_set
        candidate_spec_set = candidate_instance_set.spec_set

        logger.debug(LazyFormat(lambda: f"Candidate spec set is:\n{mf_pformat(candidate_spec_set)}"))

        data_set_linkable_specs = candidate_spec_set.linkable_specs

        # These are linkable specs in the start node data set. Those are considered "local".
        local_linkable_specs: List[LinkableInstanceSpec] = []

        # These are linkable specs that aren't in the data set, but they might be able to be joined in.
        possibly_joinable_linkable_specs: List[LinkableInstanceSpec] = []

        # Group required_linkable_specs into local / un-joinable / or possibly joinable.
        unjoinable_linkable_specs = []
        for required_linkable_spec in required_linkable_specs:
            is_metric_time = required_linkable_spec.element_name == DataSet.metric_time_dimension_name()
            is_local = required_linkable_spec in data_set_linkable_specs
            is_unjoinable = (
                # metric_time is never unjoinable. In metric queries, the agg_time_dimension is local to the measure source node.
                # In no-metric queries, can always CROSS JOIN to a time spine.
                (not is_metric_time)
                and (
                    # metric_time is the only element that can be joined without entity links.
                    len(required_linkable_spec.entity_links) == 0
                    # In order be joinable, the first entity link must be in the left node's dataset.
                    or LinklessEntitySpec.from_reference(required_linkable_spec.entity_links[0])
                    not in data_set_linkable_specs
                )
            )
            if is_local:
                local_linkable_specs.append(required_linkable_spec)
            elif is_unjoinable:
                unjoinable_linkable_specs.append(required_linkable_spec)
            else:
                possibly_joinable_linkable_specs.append(required_linkable_spec)

        candidates_for_join = self._find_joinable_candidate_nodes_that_can_satisfy_linkable_specs(
            left_node_instance_set=candidate_instance_set,
            needed_linkable_specs=possibly_joinable_linkable_specs,
            default_join_type=default_join_type,
        )
        join_candidates: List[JoinLinkableInstancesRecipe] = []

        logger.debug(LazyFormat(lambda: "Looping over nodes that can be joined to get the required linkable specs"))

        # Using a greedy approach, try to get the "possibly_joinable_linkable_specs" by iteratively joining nodes with
        # the most matching linkable specs. We try to join nodes with the most matching specs to minimize the number of
        # joins that we have to do to. A knapsack solution is ideal, but punting on that for simplicity.
        while len(possibly_joinable_linkable_specs) > 0:
            logger.debug(
                LazyFormat(lambda: f"Looking for linkable specs:\n{mf_pformat(possibly_joinable_linkable_specs)}")
            )

            # We've run out of candidate data sets, but there are more linkable specs that we need. That means the
            # rest of the linkable specs can't be joined in, and we're left with unjoinable specs remaining.
            if len(candidates_for_join) == 0:
                logger.debug(
                    LazyFormat(
                        lambda: "There are no more candidate nodes that can be joined, but not all linkable specs have "
                        "been acquired."
                    )
                )
                unjoinable_linkable_specs.extend(possibly_joinable_linkable_specs)
                break

            # Join the best candidate to realize the linkable specs
            next_candidate = candidates_for_join.pop(0)
            logger.debug(LazyFormat(lambda: f"The next candidate node to be joined is:\n{mf_pformat(next_candidate)}"))
            join_candidates.append(next_candidate)

            # Update the candidates. Since we'll be joined/ing the previously selected candidate, we no longer need
            # to get those linkable specs from the remaining candidates, and we also need to re-rank based on the new
            # requirements.
            candidates_for_join = self._update_candidates_that_can_satisfy_linkable_specs(
                candidates_for_join=candidates_for_join,
                already_satisfisfied_linkable_specs=next_candidate.satisfiable_linkable_specs,
            )

            # The once possibly joinable specs are definitely joinable and no longer need to be searched for.
            # Remove from "possibly_joinable_linkable_specs"
            possibly_joinable_linkable_specs = [
                x for x in possibly_joinable_linkable_specs if x not in next_candidate.satisfiable_linkable_specs
            ]

        logger.debug(LazyFormat(lambda: "Done evaluating possible joins"))
        return LinkableInstanceSatisfiabilityEvaluation(
            local_linkable_specs=tuple(local_linkable_specs),
            joinable_linkable_specs=tuple(
                itertools.chain.from_iterable([x.satisfiable_linkable_specs for x in join_candidates])
            ),
            join_recipes=tuple(join_candidates),
            unjoinable_linkable_specs=tuple(unjoinable_linkable_specs),
        )



from __future__ import annotations

from metricflow_semantics.naming.linkable_spec_name import DUNDER
from metricflow_semantics.specs.column_assoc import (
    ColumnAssociation,
    ColumnAssociationResolver,
)
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.instance_spec import InstanceSpec, InstanceSpecVisitor
from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.metadata_spec import MetadataSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec


class DunderColumnAssociationResolver(ColumnAssociationResolver):
    """Uses a double underscore to map specs to column names.

    For example:

    DimensionSpec(element_name='country', entity_links=['listing'])

    ->

    listing__country
    """

    def __init__(self) -> None:  # noqa: D107
        self._visitor_helper = DunderColumnAssociationResolverVisitor()

    def resolve_spec(self, spec: InstanceSpec) -> ColumnAssociation:  # noqa: D102
        return spec.accept(self._visitor_helper)


class DunderColumnAssociationResolverVisitor(InstanceSpecVisitor[ColumnAssociation]):
    """Visitor helper class for DefaultColumnAssociationResolver2."""

    def visit_metric_spec(self, metric_spec: MetricSpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(metric_spec.element_name if metric_spec.alias is None else metric_spec.alias)

    def visit_measure_spec(self, measure_spec: MeasureSpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(measure_spec.element_name)

    def visit_dimension_spec(self, dimension_spec: DimensionSpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(dimension_spec.qualified_name)

    def visit_time_dimension_spec(self, time_dimension_spec: TimeDimensionSpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(
            time_dimension_spec.qualified_name
            + (
                f"{DUNDER}{time_dimension_spec.aggregation_state.value.lower()}"
                if time_dimension_spec.aggregation_state
                else ""
            )
            + (
                f"{DUNDER}{time_dimension_spec.window_function.value.lower()}"
                if time_dimension_spec.window_function
                else ""
            )
        )

    def visit_entity_spec(self, entity_spec: EntitySpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(entity_spec.qualified_name)

    def visit_group_by_metric_spec(self, group_by_metric_spec: GroupByMetricSpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(group_by_metric_spec.qualified_name)

    def visit_metadata_spec(self, metadata_spec: MetadataSpec) -> ColumnAssociation:  # noqa: D102
        return ColumnAssociation(metadata_spec.qualified_name)



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.references import TimeDimensionReference
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class MetricTimeDimensionTransformNode(DataflowPlanNode):
    """A node transforms the input data set so that it contains the metric time dimension and relevant measures.

    The metric time dimension is used later to aggregate all measures in the data set.

    Input: a data set containing measures along with the associated aggregation time dimension.

    Output: a data set similar to the input data set, but includes the configured aggregation time dimension as the
    metric time dimension and only contains measures that are defined to use it.

    Attributes:
        aggregation_time_dimension_reference: The time dimension that measures in the input should be aggregated to.
    """

    aggregation_time_dimension_reference: TimeDimensionReference

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        aggregation_time_dimension_reference: TimeDimensionReference,
    ) -> MetricTimeDimensionTransformNode:
        return MetricTimeDimensionTransformNode(
            parent_nodes=(parent_node,),
            aggregation_time_dimension_reference=aggregation_time_dimension_reference,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_SET_MEASURE_AGGREGATION_TIME

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_metric_time_dimension_transform_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Metric Time Dimension '{self.aggregation_time_dimension_reference.element_name}'"

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("aggregation_time_dimension", self.aggregation_time_dimension_reference.element_name),
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.aggregation_time_dimension_reference == self.aggregation_time_dimension_reference
        )

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> MetricTimeDimensionTransformNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return MetricTimeDimensionTransformNode.create(
            parent_node=new_parent_nodes[0],
            aggregation_time_dimension_reference=self.aggregation_time_dimension_reference,
        )



from __future__ import annotations

import datetime
import functools
import logging
import os
import platform
import sys
import time
import traceback
import uuid
from hashlib import sha256
from typing import Callable, List, Optional, TypeVar

from metricflow_semantics.random_id import random_id
from typing_extensions import ParamSpec

from metricflow.telemetry.handlers.handlers import (
    TelemetryHandler,
    ToMemoryTelemetryHandler,
)
from metricflow.telemetry.handlers.python_log import PythonLoggerTelemetryHandler
from metricflow.telemetry.models import FunctionEndEvent, FunctionStartEvent, TelemetryLevel

logger = logging.getLogger(__name__)


class TelemetryReporter:
    """Reports telemetry for improving product experience."""

    # Session ID to use when requesting a non-uniquely identifiable ID.
    FULLY_ANONYMOUS_CLIENT_ID = "anonymous"
    ENV_EMAIL_OVERRIDE = "METRICFLOW_CLIENT_EMAIL"

    def __init__(self, report_levels_higher_or_equal_to: TelemetryLevel, fully_anonymous: bool = False) -> None:
        """If fully_anonymous is set, use a client_id that is not unique."""
        self._report_levels_higher_or_equal_to = report_levels_higher_or_equal_to
        self._fully_anonymous = fully_anonymous
        self._email = os.getenv(TelemetryReporter.ENV_EMAIL_OVERRIDE)

        if fully_anonymous:
            self._client_id = TelemetryReporter.FULLY_ANONYMOUS_CLIENT_ID
        elif self._email:
            self._client_id = self._email
        else:
            self._client_id = TelemetryReporter._create_client_id()

        # For testing
        self._test_handler = ToMemoryTelemetryHandler()
        self._handlers: List[TelemetryHandler] = []

    @staticmethod
    def _create_client_id() -> str:
        """Creates an identifier for the current user based on their current environment.

        More specifically, this function creates a SHA-256 hash based on the system platform, release, and MAC address.

        The created client ID is not guaranteed to be unique by user.
        """
        # getnode() returns the MAC.
        id_str = "_".join([sys.platform, platform.release(), str(uuid.getnode())])
        return sha256(id_str.encode("utf-8")).hexdigest()

    def add_python_log_handler(self) -> None:  # noqa: D102
        self._handlers.append(PythonLoggerTelemetryHandler(logger_level=logging.DEBUG))

    def add_test_handler(self) -> None:
        """See test_handler."""
        self._handlers.append(self._test_handler)

    @property
    def test_handler(self) -> ToMemoryTelemetryHandler:
        """Used for testing only to verify that the handlers are getting the right events."""
        return self._test_handler

    def log_function_start(
        self,
        invocation_id: str,
        module_name: str,
        function_name: str,
    ) -> None:
        """Logs the start of a function call when the logging level >= USAGE.

        invocation_id is to uniquely identify different function calls.
        """
        if TelemetryLevel.USAGE >= self._report_levels_higher_or_equal_to:
            for handler in self._handlers:
                handler.log(
                    client_id=self._client_id,
                    function_start_event=FunctionStartEvent.create(
                        event_time=datetime.datetime.now(),
                        level_name=TelemetryLevel.USAGE.name,
                        invocation_id=invocation_id,
                        module_name=module_name,
                        function_name=function_name,
                    ),
                )

    def log_function_end(
        self, invocation_id: str, module_name: str, function_name: str, runtime: float, exception_trace: Optional[str]
    ) -> None:
        """Similar to log_function_end, except adding the duration of the call and exception trace on error."""
        if TelemetryLevel.USAGE >= self._report_levels_higher_or_equal_to or (
            exception_trace and TelemetryLevel.EXCEPTION >= self._report_levels_higher_or_equal_to
        ):
            for handler in self._handlers:
                handler.log(
                    client_id=self._client_id,
                    function_end_event=FunctionEndEvent.create(
                        event_time=datetime.datetime.now(),
                        level_name=TelemetryLevel.USAGE.name if not exception_trace else TelemetryLevel.EXCEPTION.name,
                        invocation_id=invocation_id,
                        module_name=module_name,
                        function_name=function_name,
                        runtime=runtime,
                        exception_trace=exception_trace,
                    ),
                )


P = ParamSpec("P")
R = TypeVar("R")


def log_call(telemetry_reporter: TelemetryReporter, module_name: str) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator to make it easier to log telemetry for function calls.

    Using module_name instead of introspection since it seems more robust.

    Example call:

    @log_call(telemetry_reporter=telemetry_reporter, module_name=__name__)
    def test_function() -> str:
        return "foo"

    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            # Not every Callable has a __name__
            function_name = getattr(func, "__name__", repr(func))
            invocation_id = f"call_{random_id()}"
            start_time = time.time()
            telemetry_reporter.log_function_start(
                invocation_id=invocation_id, module_name=module_name, function_name=function_name
            )
            exception_trace: Optional[str] = None
            try:
                return func(*args, **kwargs)
            except Exception:
                exception_trace = traceback.format_exc()
                raise
            finally:
                telemetry_reporter.log_function_end(
                    invocation_id=invocation_id,
                    module_name=module_name,
                    function_name=function_name,
                    runtime=time.time() - start_time,
                    exception_trace=exception_trace,
                )

        return wrapped

    return decorator



from __future__ import annotations

import textwrap
from dataclasses import dataclass
from typing import Optional, Sequence

import jinja2
from dbt_semantic_interfaces.references import SemanticModelReference
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataset.sql_dataset import SqlDataSet


@dataclass(frozen=True, eq=False)
class ReadSqlSourceNode(DataflowPlanNode):
    """A source node where data from an SQL table or SQL query is read and output.

    Attributes:
        data_set: Dataset describing the SQL table / SQL query.
    """

    data_set: SqlDataSet

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 0

    @staticmethod
    def create(  # noqa: D102
        data_set: SqlDataSet,
    ) -> ReadSqlSourceNode:
        return ReadSqlSourceNode(
            parent_nodes=(),
            data_set=data_set,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_READ_SQL_SOURCE_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_source_node(self)

    @override
    @property
    def _input_semantic_model(self) -> Optional[SemanticModelReference]:
        """Return the semantic model serving as direct input for this node, if one exists."""
        return self.data_set.semantic_model_reference

    def __str__(self) -> str:  # noqa: D105
        return jinja2.Template(
            textwrap.dedent(
                """\
                <{{ class_name }} data_set={{ data_set }} />
                """
            )
        ).render(class_name=self.__class__.__name__, data_set=str(self.data_set))

    @property
    def description(self) -> str:  # noqa: D102
        return f"""Read From {self.data_set}"""

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("data_set", self.data_set),)

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__) and other_node.data_set == self.data_set

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> ReadSqlSourceNode:  # noqa: D102
        assert len(new_parent_nodes) == 0
        return ReadSqlSourceNode.create(data_set=self.data_set)



from __future__ import annotations

import datetime
import logging
from dataclasses import dataclass

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class TimeRangeConstraint(SerializableDataclass):
    """Describes how the time dimension for metrics should be constrained."""

    start_time: datetime.datetime
    end_time: datetime.datetime

    def __post_init__(self) -> None:  # noqa: D105
        if self.start_time > self.end_time:
            logger.warning(
                LazyFormat(
                    lambda: f"start_time must not be > end_time. start_time={self.start_time} end_time={self.end_time}"
                )
            )

        if self.start_time < TimeRangeConstraint.ALL_TIME_BEGIN():
            logger.warning(
                LazyFormat(
                    lambda: f"start_time={self.start_time} exceeds the limits of {TimeRangeConstraint.ALL_TIME_BEGIN()}"
                )
            )

        if self.end_time > TimeRangeConstraint.ALL_TIME_END():
            raise RuntimeError(f"end_time={self.end_time} exceeds the limits of {TimeRangeConstraint.ALL_TIME_END()}")

    @staticmethod
    def ALL_TIME_BEGIN() -> datetime.datetime:  # noqa: D102
        return datetime.datetime(2000, 1, 1)

    @staticmethod
    def ALL_TIME_END() -> datetime.datetime:  # noqa: D102
        return datetime.datetime(2040, 12, 31)

    @staticmethod
    def all_time() -> TimeRangeConstraint:
        """Return the range representing all time.

        This could also be represented with None as the ends, but doing this makes the logic simpler in many cases.
        """
        return TimeRangeConstraint(
            start_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
            end_time=TimeRangeConstraint.ALL_TIME_END(),
        )

    @staticmethod
    def empty_time() -> TimeRangeConstraint:
        """Return the range representing no time."""
        return TimeRangeConstraint(
            start_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
            end_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
        )

    def is_subset_of(self, other: TimeRangeConstraint) -> bool:  # noqa: D102
        return self.start_time >= other.start_time and self.end_time <= other.end_time

    def __str__(self) -> str:  # noqa: D105
        return f"[{self.start_time.isoformat()}, {self.end_time.isoformat()}]"

    def __repr__(self) -> str:  # noqa: D105
        return (
            f"{self.__class__.__name__}(start_time='{self.start_time.isoformat()}', "
            f"end_time='{self.end_time.isoformat()}')"
        )

    def intersection(self, other: TimeRangeConstraint) -> TimeRangeConstraint:  # noqa: D102
        # self is completely before the other
        if self.end_time < other.start_time:
            return TimeRangeConstraint.empty_time()
        # self starts before the start of other, and self ends within other
        elif self.start_time <= other.start_time <= self.end_time <= other.end_time:
            return TimeRangeConstraint(
                start_time=other.start_time,
                end_time=self.end_time,
            )
        # self starts before the start of other, and self ends after other
        elif self.start_time <= other.start_time <= other.end_time <= self.end_time:
            return other
        # self starts after the start of other, and self ends within other:
        elif other.start_time <= self.start_time <= self.end_time <= other.end_time:
            return self
        # self starts after the start of other, and self ends after other:
        elif other.start_time <= self.start_time <= other.end_time <= self.end_time:
            return TimeRangeConstraint(
                start_time=self.start_time,
                end_time=other.end_time,
            )
        # self is completely after other
        elif self.start_time > other.end_time:
            return TimeRangeConstraint.empty_time()
        else:
            raise RuntimeError(f"Unhandled case - self: {self} other: {other}")



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

from dbt_semantic_interfaces.type_enums import AggregationType
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.entity_spec import LinklessEntitySpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class SemiAdditiveJoinNode(DataflowPlanNode):
    """A node that performs a row filter by aggregating a given non-additive dimension.

    This is designed to filter a dataset down to singular non-additive time dimension values by aggregating
    the time dimension with MAX/MIN then joining back to the original dataset on that aggregated value.
    Additionally, an optional sequence of entities can be passed in to group by and join on during the filtering.

    For example, if we have a data set that includes "account_balances,user,date", we can build a
    "latest_account_balance_by_user" data set using this node by filtering the data set such that for each user,
    we get the MAX(date) and return the account_balance corresponding to that date.

    Data transformation example,
    | date       | account_balance | user |                             | date       | account_balance | user |
    |:-----------|-----------------|-----:|     entity_specs:           |:-----------|-----------------|-----:|
    | 2019-12-31 |            1000 |    u1|       - user                | 2020-01-03 |            2000 |    u1|
    | 2020-01-03 |            2000 |    u1| ->  time_dimension_spec: -> | 2020-01-12 |            1500 |    u2|
    | 2020-01-09 |            3000 |    u2|       - date                | 2020-01-12 |            1000 |    u3|
    | 2020-01-12 |            1500 |    u2|     agg_by_function:
    | 2020-01-12 |            1000 |    u3|       - MAX


    Similarly, if we don't provide any entity_specs, it would end up performing the aggregation filter only on the
    time_dimension_spec without any grouping by any entities.

    Data transformation example,
    | date       | account_balance | user |                             | date       | account_balance |
    |:-----------|-----------------|-----:|     entity_specs:           |:-----------|----------------:|
    | 2019-12-31 |            1000 |    u1|                             | 2020-01-12 |            2500 |
    | 2020-01-03 |            2000 |    u1| ->  time_dimension_spec: ->
    | 2020-01-09 |            3000 |    u2|       - date
    | 2020-01-12 |            1500 |    u2|     agg_by_function:
    | 2020-01-12 |            1000 |    u3|       - MAX


    Additionally, we can aggregate against 'windows' of a dataset. For example, if we group by a non_additive time dimension,
    we perform the MIN/MAX filtering on the granularity window provided in that group by instead of the full time range.

    Data transformation example,
    | date       | account_balance | user |                                  | date       | account_balance |
    |:-----------|-----------------|-----:|     entity_specs:                |:-----------|----------------:|
    | 2019-12-31 |            1500 |    u1|     time_dimension_spec:         | 2019-12-31 |            1500 |
    | 2020-01-03 |            2000 |    u1| ->    - date                  -> | 2020-01-07 |            3000 |
    | 2020-01-09 |            3000 |    u2|     agg_by_function:             | 2020-01-14 |            3250 |
    | 2020-01-12 |            1500 |    u2|       - MIN
    | 2020-01-14 |            1250 |    u3|     queried_time_dimension_spec:
    | 2020-01-14 |            2000 |    u2|       - date__week
    | 2020-01-15 |            4000 |    u1|

    Attributes:
        entity_specs: The entities to group the join by.
        time_dimension_spec: The time dimension used for row filtering via an aggregation.
        agg_by_function: The aggregation function used on the time dimension.
        queried_time_dimension_spec: The group by provided in the query used to build the windows we want to filter on.
    """

    entity_specs: Sequence[LinklessEntitySpec]
    time_dimension_spec: TimeDimensionSpec
    agg_by_function: AggregationType
    queried_time_dimension_spec: Optional[TimeDimensionSpec]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        entity_specs: Sequence[LinklessEntitySpec],
        time_dimension_spec: TimeDimensionSpec,
        agg_by_function: AggregationType,
        queried_time_dimension_spec: Optional[TimeDimensionSpec] = None,
    ) -> SemiAdditiveJoinNode:
        return SemiAdditiveJoinNode(
            parent_nodes=(parent_node,),
            entity_specs=tuple(entity_specs),
            time_dimension_spec=time_dimension_spec,
            agg_by_function=agg_by_function,
            queried_time_dimension_spec=queried_time_dimension_spec,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_SEMI_ADDITIVE_JOIN_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_semi_additive_join_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"""Join on {self.agg_by_function.name}({self.time_dimension_spec.element_name}) and {[i.element_name for i in self.entity_specs]} grouping by {self.queried_time_dimension_spec.element_name if self.queried_time_dimension_spec else None}"""

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return super().displayed_properties

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        if not isinstance(other_node, self.__class__):
            return False

        return (
            other_node.entity_specs == self.entity_specs
            and other_node.time_dimension_spec == self.time_dimension_spec
            and other_node.agg_by_function == self.agg_by_function
            and other_node.queried_time_dimension_spec == self.queried_time_dimension_spec
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> SemiAdditiveJoinNode:  # noqa: D102
        assert len(new_parent_nodes) == 1

        return SemiAdditiveJoinNode.create(
            parent_node=new_parent_nodes[0],
            entity_specs=self.entity_specs,
            time_dimension_spec=self.time_dimension_spec,
            agg_by_function=self.agg_by_function,
            queried_time_dimension_spec=self.queried_time_dimension_spec,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime


class TimeSource(ABC):
    """Provides time to classes that need a sense of time.

    A static time source can be used for testing, while a time source that uses the current time is used for production.
    """

    @abstractmethod
    def get_time(self) -> datetime:  # noqa: D102
        pass



from __future__ import annotations

import logging

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

from metricflow.telemetry.handlers.handlers import PayloadType, TelemetryHandler

logger = logging.getLogger(__name__)


class PythonLoggerTelemetryHandler(TelemetryHandler):
    """A telemetry client that logs data to the Python logger for debugging during tests."""

    def __init__(self, logger_level: int) -> None:  # noqa: D107
        self._logger_level = logger_level

    def _write_log(self, client_id: str, payload: PayloadType) -> None:
        logger.log(level=self._logger_level, msg=LazyFormat("Logging telemetry payload", payload=payload))



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence, Tuple, Union

from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.query.group_by_item.resolution_dag.input_metric_location import InputMetricDefinitionLocation
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import (
    GroupByItemResolutionNode,
    GroupByItemResolutionNodeSet,
    GroupByItemResolutionNodeVisitor,
)
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.measure_source_node import (
    MeasureGroupByItemSourceNode,
)
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class MetricGroupByItemResolutionNode(GroupByItemResolutionNode):
    """Outputs group-by-items relevant to a metric based on the input group-by-items.

    Attributes:
        metric_reference: The metric that this represents.
        metric_input_location: If this is an input metric for a derived metric, the location within the derived metric definition.
        parent_nodes: The parent nodes of this metric.
    """

    metric_reference: MetricReference
    metric_input_location: Optional[InputMetricDefinitionLocation]
    parent_nodes: Tuple[Union[MeasureGroupByItemSourceNode, MetricGroupByItemResolutionNode], ...]

    @staticmethod
    def create(  # noqa: D102
        metric_reference: MetricReference,
        metric_input_location: Optional[InputMetricDefinitionLocation],
        parent_nodes: Sequence[Union[MeasureGroupByItemSourceNode, MetricGroupByItemResolutionNode]],
    ) -> MetricGroupByItemResolutionNode:
        return MetricGroupByItemResolutionNode(
            metric_reference=metric_reference,
            metric_input_location=metric_input_location,
            parent_nodes=tuple(parent_nodes),
        )

    @override
    def accept(self, visitor: GroupByItemResolutionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        return visitor.visit_metric_node(self)

    @property
    @override
    def description(self) -> str:
        return "Output group-by-items available for this metric."

    @classmethod
    @override
    def id_prefix(cls) -> IdPrefix:
        return StaticIdPrefix.METRIC_GROUP_BY_ITEM_RESOLUTION_NODE

    @property
    @override
    def displayed_properties(self) -> Sequence[DisplayedProperty]:
        return tuple(super().displayed_properties) + (
            DisplayedProperty(
                key="metric_reference",
                value=str(self.metric_reference),
            ),
        )

    @property
    @override
    def ui_description(self) -> str:
        if self.metric_input_location is None:
            return f"Metric({repr(self.metric_reference.element_name)})"
        return (
            f"Metric({repr(self.metric_reference.element_name)}, "
            f"input_metric_index={self.metric_input_location.input_metric_list_index})"
        )

    @override
    def _self_set(self) -> GroupByItemResolutionNodeSet:
        return GroupByItemResolutionNodeSet(metric_nodes=(self,))



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import (
    ResolverInputForOrderByItem,
)


@dataclass(frozen=True)
class InvalidOrderByItemIssue(MetricFlowQueryResolutionIssue):
    """Describes a query issue where the order-by item does not match one of the queried metrics / group-by-items."""

    order_by_item_input: ResolverInputForOrderByItem

    @staticmethod
    def from_parameters(  # noqa: D102
        order_by_item_input: ResolverInputForOrderByItem,
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> InvalidOrderByItemIssue:
        return InvalidOrderByItemIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            order_by_item_input=order_by_item_input,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return (
            f"The order-by item {repr(self.order_by_item_input.input_obj)} does not match exactly one "
            f"of the query items."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> InvalidOrderByItemIssue:
        return InvalidOrderByItemIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            order_by_item_input=self.order_by_item_input,
        )



from __future__ import annotations

import functools
from abc import ABC, abstractmethod
from typing import Iterable, Type, TypeVar

from typing_extensions import Self

MergeableT = TypeVar("MergeableT", bound="Mergeable")


class Mergeable(ABC):
    """Objects that can be merged together to form a superset object of the same type.

    Merging objects are frequently needed in MetricFlow as there are several recursive operations where the output is
    the superset of the result of the recursive calls.

    e.g.
    * The validation issue set of a derived metric includes the issues of the parent metrics.
    * The output of a node in the dataflow plan can include the outputs of the parent nodes.
    * The query-time where filter is useful to combine with the metric-defined where filter.

    Having a common interface also gives a consistent name to this operation so that we don't end up with multiple names
    to describe this operation (e.g. combine, add, concat).

    This is used to streamline the following case that occurs in the codebase:

        items_to_merge: List[ItemType] = []
        ...
        if ...
            items_to_merge.append(...)
        ...
        if ...
            items_to_merge.append(...)
        ...
        if ...
            ...
            items_to_merge.append(...)
            return merge_items(items_to_merge)
        ...
        return merge_items(items_to_merge)

    This centralizes the definition of the merge_items() call.
    """

    @abstractmethod
    def merge(self: Self, other: Self) -> Self:
        """Return a new object that is the result of merging self with other."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def empty_instance(cls: Type[MergeableT]) -> MergeableT:
        """Create an empty instance to handle merging of empty sequences of items.

        As merge_iterable() returns an empty instance for an empty iterable, there needs to be a way of creating one.
        """
        raise NotImplementedError

    @classmethod
    def merge_iterable(cls: Type[MergeableT], items: Iterable[MergeableT]) -> MergeableT:
        """Merge all items into a single instance.

        If an empty iterable has been passed in, this returns an empty instance.
        """
        return functools.reduce(cls.merge, items, cls.empty_instance())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import EntityReference
from typing_extensions import override

from metricflow_semantics.model.semantics.linkable_element import ElementPathKey, LinkableElementType
from metricflow_semantics.specs.instance_spec import InstanceSpecVisitor, LinkableInstanceSpec
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class EntitySpec(LinkableInstanceSpec, SerializableDataclass):  # noqa: D101
    @property
    def without_first_entity_link(self) -> EntitySpec:  # noqa: D102
        assert len(self.entity_links) > 0, f"Spec does not have any entity links: {self}"
        return EntitySpec(element_name=self.element_name, entity_links=self.entity_links[1:])

    @property
    def without_entity_links(self) -> EntitySpec:  # noqa: D102
        return LinklessEntitySpec.from_element_name(self.element_name)

    @property
    def as_linkless_prefix(self) -> Tuple[EntityReference, ...]:
        """Creates tuple of linkless entities that could be included in the entity_links of another spec.

        eg as a prefix to a DimensionSpec's entity links to when a join is occurring via this entity
        """
        return (EntityReference(element_name=self.element_name),) + self.entity_links

    def __eq__(self, other: Any) -> bool:  # type: ignore[misc]  # noqa: D105
        if not isinstance(other, EntitySpec):
            return False
        return self.element_name == other.element_name and self.entity_links == other.entity_links

    def __hash__(self) -> int:  # noqa: D105
        return hash((self.element_name, self.entity_links))

    @property
    def reference(self) -> EntityReference:  # noqa: D102
        return EntityReference(element_name=self.element_name)

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_entity_spec(self)

    @property
    @override
    def element_path_key(self) -> ElementPathKey:
        return ElementPathKey(
            element_name=self.element_name, element_type=LinkableElementType.ENTITY, entity_links=self.entity_links
        )

    def with_entity_prefix(self, entity_prefix: EntityReference) -> EntitySpec:  # noqa: D102
        return EntitySpec(element_name=self.element_name, entity_links=(entity_prefix,) + self.entity_links)


@dataclass(frozen=True)
class LinklessEntitySpec(EntitySpec, SerializableDataclass):
    """Similar to EntitySpec, but requires that it doesn't have entity links."""

    @staticmethod
    def from_element_name(element_name: str) -> LinklessEntitySpec:  # noqa: D102
        return LinklessEntitySpec(element_name=element_name, entity_links=())

    def __post_init__(self) -> None:  # noqa: D105
        if len(self.entity_links) > 0:
            raise RuntimeError(f"{self.__class__.__name__} shouldn't have entity links. Got: {self}")

    def __eq__(self, other: Any) -> bool:  # type: ignore[misc]  # noqa: D105
        if not isinstance(other, EntitySpec):
            return False
        return self.element_name == other.element_name and self.entity_links == other.entity_links

    def __hash__(self) -> int:  # noqa: D105
        return hash((self.element_name, self.entity_links))

    @staticmethod
    def from_reference(entity_reference: EntityReference) -> LinklessEntitySpec:  # noqa: D102
        return LinklessEntitySpec(element_name=entity_reference.element_name, entity_links=())



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Sequence

from dbt_semantic_interfaces.protocols import SemanticModel
from dbt_semantic_interfaces.references import DimensionReference
from dbt_semantic_interfaces.type_enums import DimensionType

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat


@dataclass(frozen=True)
class DimensionInvariant:
    """For a given manifest, all defined dimensions with the same name should have these same properties."""

    dimension_type: DimensionType
    is_partition: bool


class DimensionLookup:
    """Looks up properties related to dimensions."""

    def __init__(self, semantic_models: Sequence[SemanticModel]) -> None:  # noqa: D107
        self._dimension_reference_to_invariant: Dict[DimensionReference, DimensionInvariant] = {}
        for semantic_model in semantic_models:
            for dimension in semantic_model.dimensions:
                invariant = DimensionInvariant(
                    dimension_type=dimension.type,
                    is_partition=dimension.is_partition,
                )
                dimension_reference = dimension.reference
                existing_invariant = self._dimension_reference_to_invariant.get(dimension_reference)
                if existing_invariant is not None and existing_invariant != invariant:
                    raise ValueError(
                        str(
                            LazyFormat(
                                "Dimensions with the same name have been defined with conflicting values that "
                                "should have been the same in a given semantic manifest. This should have been caught "
                                "during validation.",
                                dimension_reference=dimension_reference,
                                existing_invariant=existing_invariant,
                                conflicting_invariant=invariant,
                                semantic_model_reference=semantic_model.reference,
                            )
                        )
                    )

                self._dimension_reference_to_invariant[dimension_reference] = invariant

    def get_invariant(self, dimension_reference: DimensionReference) -> DimensionInvariant:
        """Get invariants for the given dimension in the semantic manifest."""
        # dimension_reference might be a TimeDimensionReference, so change types.
        dimension_reference = DimensionReference(element_name=dimension_reference.element_name)
        invariant = self._dimension_reference_to_invariant[dimension_reference]
        if invariant is None:
            raise ValueError(
                str(
                    LazyFormat(
                        "Unknown dimension reference",
                        dimension_reference=dimension_reference,
                        known_dimension_references=list(self._dimension_reference_to_invariant.keys()),
                    )
                )
            )

        return invariant



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from metricflow_semantics.collection_helpers.lru_cache import LruCache
from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.where_filter.where_filter_transform import WhereSpecFactory

from metricflow.dataflow.builder.measure_spec_properties import MeasureSpecProperties
from metricflow.dataflow.builder.source_node_recipe import SourceNodeRecipe
from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.plan_conversion.node_processor import PredicatePushdownState


@dataclass(frozen=True)
class FindSourceNodeRecipeParameterSet:
    """Parameters for `DataflowPlanBuilder._find_source_node_recipe()`."""

    linkable_spec_set: LinkableSpecSet
    predicate_pushdown_state: PredicatePushdownState
    measure_spec_properties: Optional[MeasureSpecProperties]


@dataclass(frozen=True)
class FindSourceNodeRecipeResult:
    """Result for `DataflowPlanBuilder._find_source_node_recipe()`."""

    source_node_recipe: Optional[SourceNodeRecipe]


@dataclass(frozen=True)
class BuildAnyMetricOutputNodeParameterSet:
    """Parameters for `DataflowPlanBuilder._build_any_metric_output_node()`."""

    metric_spec: MetricSpec
    queried_linkable_specs: LinkableSpecSet
    filter_spec_factory: WhereSpecFactory
    predicate_pushdown_state: PredicatePushdownState
    for_group_by_source_node: bool


class DataflowPlanBuilderCache:
    """Cache for internal methods in `DataflowPlanBuilder`."""

    def __init__(  # noqa: D107
        self, find_source_node_recipe_cache_size: int = 1000, build_any_metric_output_node_cache_size: int = 1000
    ) -> None:
        self._find_source_node_recipe_cache = LruCache[FindSourceNodeRecipeParameterSet, FindSourceNodeRecipeResult](
            find_source_node_recipe_cache_size
        )
        self._build_any_metric_output_node_cache = LruCache[BuildAnyMetricOutputNodeParameterSet, DataflowPlanNode](
            build_any_metric_output_node_cache_size
        )

        assert find_source_node_recipe_cache_size > 0
        assert build_any_metric_output_node_cache_size > 0

    def get_find_source_node_recipe_result(  # noqa: D102
        self, parameter_set: FindSourceNodeRecipeParameterSet
    ) -> Optional[FindSourceNodeRecipeResult]:
        return self._find_source_node_recipe_cache.get(parameter_set)

    def set_find_source_node_recipe_result(  # noqa: D102
        self, parameter_set: FindSourceNodeRecipeParameterSet, source_node_recipe: FindSourceNodeRecipeResult
    ) -> None:
        self._find_source_node_recipe_cache.set(parameter_set, source_node_recipe)

    def get_build_any_metric_output_node_result(  # noqa: D102
        self, parameter_set: BuildAnyMetricOutputNodeParameterSet
    ) -> Optional[DataflowPlanNode]:
        return self._build_any_metric_output_node_cache.get(parameter_set)

    def set_build_any_metric_output_node_result(  # noqa: D102
        self, parameter_set: BuildAnyMetricOutputNodeParameterSet, dataflow_plan_node: DataflowPlanNode
    ) -> None:
        self._build_any_metric_output_node_cache.set(parameter_set, dataflow_plan_node)



from __future__ import annotations

from typing import Dict, Mapping, Sequence

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols import Dimension
from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.measure import Measure
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    EntityReference,
    LinkableElementReference,
    MeasureReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums import DimensionType, EntityType, TimeGranularity


class SemanticModelHelper:
    """Static helper methods for retrieving items from a semantic model."""

    @staticmethod
    def get_entity_from_semantic_model(
        semantic_model: SemanticModel, entity_reference: LinkableElementReference
    ) -> Entity:
        """Get entity from semantic model."""
        for entity in semantic_model.entities:
            if entity.reference == entity_reference:
                return entity

        raise ValueError(
            f"No entity with name ({entity_reference}) in semantic_model with name ({semantic_model.name})"
        )

    @staticmethod
    def resolved_primary_entity(semantic_model: SemanticModel) -> EntityReference:
        """Return the primary entity for dimensions in the model.

        Semantic models with measures or dimensions should have a primary entity as enforced by semantic manifest
        validation.
        """
        primary_entity_reference = semantic_model.primary_entity_reference

        entities_with_type_primary = tuple(
            entity for entity in semantic_model.entities if entity.type == EntityType.PRIMARY
        )

        # This should be caught by the validation, but adding a sanity check.
        assert len(entities_with_type_primary) <= 1, f"Found > 1 primary entity in {semantic_model}"
        if primary_entity_reference is not None:
            assert len(entities_with_type_primary) == 0, (
                f"The primary_entity field was set to {primary_entity_reference}, but there are non-zero entities with "
                f"type {EntityType.PRIMARY} in {semantic_model}"
            )
            return primary_entity_reference

        if len(entities_with_type_primary) > 0:
            return entities_with_type_primary[0].reference

        raise ValueError(f"No primary entity found in {semantic_model.reference=}")

    @staticmethod
    def entity_links_for_local_elements(semantic_model: SemanticModel) -> Sequence[EntityReference]:
        """Return the entity prefix that can be used to access dimensions defined in the semantic model."""
        primary_entity_reference = semantic_model.primary_entity_reference

        possible_entity_links = set()
        if primary_entity_reference is not None:
            possible_entity_links.add(primary_entity_reference)

        for entity in semantic_model.entities:
            if entity.is_linkable_entity_type:
                possible_entity_links.add(entity.reference)

        return sorted(possible_entity_links, key=lambda entity_reference: entity_reference.element_name)

    @staticmethod
    def get_measure_from_semantic_model(semantic_model: SemanticModel, measure_reference: MeasureReference) -> Measure:
        """Get measure from semantic model."""
        for measure in semantic_model.measures:
            if measure.reference == measure_reference:
                return measure

        raise ValueError(
            f"No dimension with name ({measure_reference.element_name}) in semantic_model with name ({semantic_model.name})"
        )

    @staticmethod
    def get_dimension_from_semantic_model(
        semantic_model: SemanticModel, dimension_reference: LinkableElementReference
    ) -> Dimension:
        """Get dimension from semantic model."""
        for dim in semantic_model.dimensions:
            if dim.reference == dimension_reference:
                return dim
        raise ValueError(
            f"No dimension with name ({dimension_reference}) in semantic_model with name ({semantic_model.name})"
        )

    @staticmethod
    def get_time_dimension_grains(semantic_model: SemanticModel) -> Mapping[TimeDimensionReference, TimeGranularity]:
        """Return a mapping of the defined time granularity of the time dimensions in the semantic mode."""
        time_dimension_reference_to_grain: Dict[TimeDimensionReference, TimeGranularity] = {}

        for dimension in semantic_model.dimensions:
            if dimension.type is DimensionType.TIME:
                if dimension.type_params is None:
                    raise ValueError(
                        f"A dimension is specified as a time dimension but does not specify a gain. This should have "
                        f"been caught in semantic-manifest validation {dimension=} {semantic_model=}"
                    )
                time_dimension_reference_to_grain[
                    dimension.reference.time_dimension_reference
                ] = dimension.type_params.time_granularity
            elif dimension.type is DimensionType.CATEGORICAL:
                pass
            else:
                assert_values_exhausted(dimension.type)

        return time_dimension_reference_to_grain



from __future__ import annotations

import logging
import time
from typing import Dict, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols.metric import Metric, MetricInputMeasure, MetricType
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.references import MeasureReference, MetricReference
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.collection_helpers.lru_cache import LruCache
from metricflow_semantics.errors.error_classes import DuplicateMetricError, MetricNotFoundError, NonExistentMeasureError
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.model.semantics.linkable_spec_resolver import (
    ValidLinkableSpecResolver,
)
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import MAX_JOIN_HOPS
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


class MetricLookup:
    """Tracks semantic information for metrics by linking them to semantic models."""

    def __init__(
        self,
        semantic_manifest: SemanticManifest,
        semantic_model_lookup: SemanticModelLookup,
        custom_granularities: Dict[str, ExpandedTimeGranularity],
    ) -> None:
        """Initializer.

        Args:
            semantic_manifest: used to fetch and load the metrics and initialize the linkable spec resolver
            semantic_model_lookup: provides access to semantic model metadata for various lookup operations
        """
        self._metrics: Dict[MetricReference, Metric] = {}
        self._semantic_model_lookup = semantic_model_lookup
        self._custom_granularities = custom_granularities

        for metric in semantic_manifest.metrics:
            self._add_metric(metric)

        self._linkable_spec_resolver = ValidLinkableSpecResolver(
            semantic_manifest=semantic_manifest,
            semantic_model_lookup=semantic_model_lookup,
            max_entity_links=MAX_JOIN_HOPS,
        )

        # Cache for `get_min_queryable_time_granularity()`
        self._metric_reference_to_min_metric_time_grain: Dict[MetricReference, TimeGranularity] = {}

        # Cache for `get_valid_agg_time_dimensions_for_metric()`.
        self._metric_reference_to_valid_agg_time_dimension_specs: Dict[
            MetricReference, Sequence[TimeDimensionSpec]
        ] = {}

        # Cache for `linkable_elements_for_measure()`.
        self._linkable_element_set_for_measure_cache: Dict[
            Tuple[MeasureReference, LinkableElementFilter], LinkableElementSet
        ] = {}

        self._linkable_elements_including_group_by_metrics_cache = LruCache[
            Tuple[MeasureReference, LinkableElementFilter], LinkableElementSet
        ](128)
        self._linkable_elements_for_no_metrics_query_cache = LruCache[LinkableElementFilter, LinkableElementSet](128)
        self._linkable_elements_for_metrics_cache = LruCache[
            Tuple[Sequence[MetricReference], LinkableElementFilter], LinkableElementSet
        ](128)

    def linkable_elements_for_measure(
        self,
        measure_reference: MeasureReference,
        element_filter: LinkableElementFilter = LinkableElementFilter(),
    ) -> LinkableElementSet:
        """Return the set of linkable elements reachable from a given measure."""
        start_time = time.time()

        # Cache the result when group-by-metrics are selected in an LRU cache as there can be many of them and may
        # significantly increase memory usage.
        if (
            LinkableElementProperty.METRIC in element_filter.with_any_of
            and LinkableElementProperty.METRIC not in element_filter.without_any_of
        ):
            cache_key = (measure_reference, element_filter)
            result = self._linkable_elements_including_group_by_metrics_cache.get(cache_key)
            if result is not None:
                return result

            result = self._linkable_spec_resolver.get_linkable_element_set_for_measure(
                measure_reference, element_filter
            )
            self._linkable_elements_including_group_by_metrics_cache.set(cache_key, result)
            return result

        # Cache the result without element names in the filter for better hit rates.
        element_filter_without_element_names = element_filter.without_element_names()
        cache_key = (measure_reference, element_filter_without_element_names)

        result = self._linkable_element_set_for_measure_cache.get(cache_key)
        if result is not None:
            return result.filter(element_filter)

        result = self._linkable_spec_resolver.get_linkable_element_set_for_measure(
            measure_reference, element_filter_without_element_names
        )
        self._linkable_element_set_for_measure_cache[cache_key] = result

        logger.debug(
            LazyFormat(
                "Finished getting linkable elements",
                measure_reference=measure_reference,
                element_filter=element_filter,
                runtime=time.time() - start_time,
            )
        )
        return result.filter(element_filter)

    def linkable_elements_for_no_metrics_query(
        self, element_set_filter: LinkableElementFilter = LinkableElementFilter()
    ) -> LinkableElementSet:
        """Return the reachable linkable elements for a dimension values query with no metrics."""
        cache_key = element_set_filter
        result = self._linkable_elements_for_no_metrics_query_cache.get(cache_key)
        if result is not None:
            return result

        result = self._linkable_spec_resolver.get_linkable_elements_for_distinct_values_query(element_set_filter)
        self._linkable_elements_for_no_metrics_query_cache.set(cache_key, result)
        return result

    def linkable_elements_for_metrics(
        self, metric_references: Sequence[MetricReference], element_set_filter: LinkableElementFilter
    ) -> LinkableElementSet:
        """Retrieve the matching set of linkable elements common to all metrics requested (intersection)."""
        cache_key = (metric_references, element_set_filter)
        result = self._linkable_elements_for_metrics_cache.get(cache_key)
        if result is not None:
            return result

        result = self._linkable_spec_resolver.get_linkable_elements_for_metrics(
            metric_references=metric_references, element_filter=element_set_filter
        )
        self._linkable_elements_for_metrics_cache.set(cache_key, result)
        return result

    def get_metrics(self, metric_references: Sequence[MetricReference]) -> Sequence[Metric]:  # noqa: D102
        res = []
        for metric_reference in metric_references:
            if metric_reference not in self._metrics:
                raise MetricNotFoundError(
                    f"Unable to find metric `{metric_reference}`. Perhaps it has not been registered"
                )
            res.append(self._metrics[metric_reference])

        return res

    @property
    def metric_references(self) -> Sequence[MetricReference]:  # noqa: D102
        return list(self._metrics.keys())

    def get_metric(self, metric_reference: MetricReference) -> Metric:  # noqa: D102
        if metric_reference not in self._metrics:
            raise MetricNotFoundError(f"Unable to find metric `{metric_reference}`. Perhaps it has not been registered")
        return self._metrics[metric_reference]

    def _add_metric(self, metric: Metric) -> None:
        """Add metric, validating presence of required measures."""
        metric_reference = MetricReference(element_name=metric.name)
        if metric_reference in self._metrics:
            raise DuplicateMetricError(f"Metric `{metric.name}` has already been registered")
        for measure_reference in metric.measure_references:
            if measure_reference not in self._semantic_model_lookup.measure_references:
                raise NonExistentMeasureError(
                    f"Metric `{metric.name}` references measure `{measure_reference}` which has not been registered"
                )
        self._metrics[metric_reference] = metric

    def configured_input_measure_for_metric(  # noqa: D102
        self, metric_reference: MetricReference
    ) -> Optional[MetricInputMeasure]:  # noqa: D102
        metric = self.get_metric(metric_reference=metric_reference)
        if metric.type is MetricType.CUMULATIVE or metric.type is MetricType.SIMPLE:
            assert len(metric.input_measures) == 1, "Simple and cumulative metrics should have one input measure."
            return metric.input_measures[0]
        elif (
            metric.type is MetricType.RATIO or metric.type is MetricType.DERIVED or metric.type is MetricType.CONVERSION
        ):
            return None
        else:
            assert_values_exhausted(metric.type)

    def contains_cumulative_or_time_offset_metric(self, metric_references: Sequence[MetricReference]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric or a derived metric with time offset."""
        for metric_reference in metric_references:
            metric = self.get_metric(metric_reference)
            if metric.type == MetricType.CUMULATIVE:
                return True
            elif metric.type == MetricType.DERIVED:
                for input_metric in metric.type_params.metrics or []:
                    if input_metric.offset_window or input_metric.offset_to_grain:
                        return True
        return False

    def _get_agg_time_dimension_specs_for_metric(
        self, metric_reference: MetricReference
    ) -> Sequence[TimeDimensionSpec]:
        """Retrieves the aggregate time dimensions associated with the metric's measures."""
        metric = self.get_metric(metric_reference)
        specs: Set[TimeDimensionSpec] = set()
        for input_measure in metric.input_measures:
            measure_properties = self._semantic_model_lookup.measure_lookup.get_properties(
                measure_reference=input_measure.measure_reference
            )
            specs.update(measure_properties.agg_time_dimension_specs)
        return list(specs)

    def get_valid_agg_time_dimensions_for_metric(
        self, metric_reference: MetricReference
    ) -> Sequence[TimeDimensionSpec]:
        """Get the agg time dimension specs that can be used in place of metric time for this metric, if applicable."""
        result = self._metric_reference_to_valid_agg_time_dimension_specs.get(metric_reference)
        if result is not None:
            return result

        result = self._get_valid_agg_time_dimensions_for_metric(metric_reference)
        self._metric_reference_to_valid_agg_time_dimension_specs[metric_reference] = result

        return result

    def _get_valid_agg_time_dimensions_for_metric(
        self, metric_reference: MetricReference
    ) -> Sequence[TimeDimensionSpec]:
        agg_time_dimension_specs = self._get_agg_time_dimension_specs_for_metric(metric_reference)
        distinct_agg_time_dimension_identifiers = set(
            [(spec.reference, spec.entity_links) for spec in agg_time_dimension_specs]
        )
        if len(distinct_agg_time_dimension_identifiers) != 1:
            # If the metric's input measures have different agg_time_dimensions, user must use metric_time.
            return []

        agg_time_dimension_reference, agg_time_dimension_entity_links = distinct_agg_time_dimension_identifiers.pop()
        valid_agg_time_dimension_specs = TimeDimensionSpec.generate_possible_specs_for_time_dimension(
            time_dimension_reference=agg_time_dimension_reference,
            entity_links=agg_time_dimension_entity_links,
            custom_granularities=self._custom_granularities,
        )
        return valid_agg_time_dimension_specs

    def get_min_queryable_time_granularity(self, metric_reference: MetricReference) -> TimeGranularity:
        """The minimum grain that can be queried with this metric.

        Maps to the largest granularity defined for any of the metric's agg_time_dimensions.
        """
        result = self._metric_reference_to_min_metric_time_grain.get(metric_reference)
        if result is not None:
            return result

        result = self._get_min_queryable_time_granularity(metric_reference)
        self._metric_reference_to_min_metric_time_grain[metric_reference] = result
        return result

    def _get_min_queryable_time_granularity(self, metric_reference: MetricReference) -> TimeGranularity:
        metric = self.get_metric(metric_reference)
        agg_time_dimension_grains = set()
        for input_measure in metric.input_measures:
            measure_properties = self._semantic_model_lookup.measure_lookup.get_properties(
                input_measure.measure_reference
            )
            agg_time_dimension_grains.add(measure_properties.agg_time_granularity)

        return max(agg_time_dimension_grains, key=lambda time_granularity: time_granularity.to_int())

    def get_joinable_scd_specs_for_metric(self, metric_reference: MetricReference) -> Sequence[LinkableInstanceSpec]:
        """Get the SCDs that can be joined to a metric."""
        filter = LinkableElementFilter(
            with_any_of=frozenset([LinkableElementProperty.SCD_HOP]),
        )
        scd_elems = self.linkable_elements_for_metrics(
            metric_references=(metric_reference,),
            element_set_filter=filter,
        )

        return scd_elems.specs



from __future__ import annotations

import re
from typing import Optional, Sequence, Tuple

from dbt_semantic_interfaces.naming.keywords import DUNDER
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.type_enums import TimeGranularity
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.naming.naming_scheme import QueryItemNamingScheme
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.patterns.entity_link_pattern import (
    EntityLinkPattern,
    ParameterSetField,
    SpecPatternParameterSet,
)
from metricflow_semantics.specs.spec_set import InstanceSpecSet, InstanceSpecSetTransform, group_spec_by_type


class DunderNamingScheme(QueryItemNamingScheme):
    """A naming scheme using the dundered name syntax.

    TODO: Consolidate with StructuredLinkableSpecName / DunderedNameFormatter.
    """

    _INPUT_REGEX = re.compile(r"\A[a-z]([a-z0-9_])*[a-z0-9]\Z")

    @staticmethod
    def date_part_suffix(date_part: DatePart) -> str:
        """Suffix used for names with a date_part."""
        return f"extract_{date_part.value}"

    @override
    def input_str(self, instance_spec: InstanceSpec) -> Optional[str]:
        spec_set = group_spec_by_type(instance_spec)

        for time_dimension_spec in spec_set.time_dimension_specs:
            # From existing comment in StructuredLinkableSpecName:
            #
            # Dunder syntax not supported for querying date_part
            #
            if time_dimension_spec.date_part is not None:
                return None
        names = _DunderNameTransform().transform(spec_set)
        if len(names) != 1:
            raise RuntimeError(f"Did not get 1 name for {instance_spec}. Got {names}")

        return names[0]

    @override
    def spec_pattern(self, input_str: str, semantic_manifest_lookup: SemanticManifestLookup) -> EntityLinkPattern:
        if not self.input_str_follows_scheme(input_str):
            raise ValueError(f"{repr(input_str)} does not follow this scheme.")

        input_str = input_str.lower()

        input_str_parts = input_str.split(DUNDER)
        fields_to_compare: Tuple[ParameterSetField, ...] = (
            ParameterSetField.ELEMENT_NAME,
            ParameterSetField.ENTITY_LINKS,
            ParameterSetField.DATE_PART,
        )

        # No dunder, e.g. "ds"
        if len(input_str_parts) == 1:
            return EntityLinkPattern(
                parameter_set=SpecPatternParameterSet.from_parameters(
                    element_name=input_str_parts[0],
                    entity_links=(),
                    time_granularity_name=None,
                    date_part=None,
                    fields_to_compare=tuple(fields_to_compare),
                )
            )

        # At this point, len(input_str_parts) >= 2
        valid_granularity_names = {granularity.value for granularity in TimeGranularity}.union(
            set(semantic_manifest_lookup.custom_granularities.keys())
        )
        suffix = input_str_parts[-1]
        time_granularity_name = suffix if suffix in valid_granularity_names else None
        # Has a time grain specified.
        if time_granularity_name is not None:
            fields_to_compare = fields_to_compare + (ParameterSetField.TIME_GRANULARITY,)
            #  e.g. "ds__month"
            if len(input_str_parts) == 2:
                return EntityLinkPattern(
                    parameter_set=SpecPatternParameterSet.from_parameters(
                        element_name=input_str_parts[0],
                        entity_links=(),
                        time_granularity_name=time_granularity_name,
                        date_part=None,
                        fields_to_compare=fields_to_compare,
                    )
                )
            # e.g. "messages__ds__month"
            return EntityLinkPattern(
                parameter_set=SpecPatternParameterSet.from_parameters(
                    element_name=input_str_parts[-2],
                    entity_links=tuple(EntityReference(entity_name) for entity_name in input_str_parts[:-2]),
                    time_granularity_name=time_granularity_name,
                    date_part=None,
                    fields_to_compare=fields_to_compare,
                )
            )

        # e.g. "messages__ds"
        return EntityLinkPattern(
            parameter_set=SpecPatternParameterSet.from_parameters(
                element_name=suffix,
                entity_links=tuple(EntityReference(entity_name) for entity_name in input_str_parts[:-1]),
                time_granularity_name=None,
                date_part=None,
                fields_to_compare=fields_to_compare,
            )
        )

    @override
    def input_str_follows_scheme(self, input_str: str) -> bool:
        # This naming scheme is case-insensitive.
        input_str = input_str.lower()
        if DunderNamingScheme._INPUT_REGEX.match(input_str) is None:
            return False

        input_str_parts = input_str.split(DUNDER)

        for date_part in DatePart:
            # TODO: We don't enforce uniqueness against date part names. Could be blocking valid dimension names.
            if input_str_parts[-1] == DunderNamingScheme.date_part_suffix(date_part=date_part):
                # From existing message in StructuredLinkableSpecName: "Dunder syntax not supported for querying
                # date_part".
                return False

        return True

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id()={hex(id(self))})"


class _DunderNameTransform(InstanceSpecSetTransform[Sequence[str]]):
    """Transforms group-by-item spec into the dundered name."""

    @override
    def transform(self, spec_set: InstanceSpecSet) -> Sequence[str]:
        names_to_return = []

        for time_dimension_spec in spec_set.time_dimension_specs:
            items = list(entity_link.element_name for entity_link in time_dimension_spec.entity_links) + [
                time_dimension_spec.element_name
            ]
            if time_dimension_spec.date_part is not None:
                items.append(DunderNamingScheme.date_part_suffix(date_part=time_dimension_spec.date_part))
            else:
                items.append(time_dimension_spec.time_granularity.name)
            names_to_return.append(DUNDER.join(items))

        for other_group_by_item_specs in (
            spec_set.entity_specs + spec_set.dimension_specs + spec_set.group_by_metric_specs
        ):
            items = list(entity_link.element_name for entity_link in other_group_by_item_specs.entity_links) + [
                other_group_by_item_specs.element_name
            ]
            names_to_return.append(DUNDER.join(items))

        return sorted(names_to_return)
